""" FreeDiskSpacePolicy

   FreeDiskSpacePolicy.__bases__:
     DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase.PolicyBase

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase


class FreeDiskSpacePolicy(PolicyBase):
    """
    The FreeDiskSpacePolicy class is a policy class satisfied when a SE has a
    low occupancy.

    FreeDiskSpacePolicy, given the space left at the element, proposes a new status.
    """

    @staticmethod
    def _evaluate(commandResult):
        """
        Evaluate policy on SE occupancy: Use FreeDiskSpaceCommand

        :Parameters:
          **commandResult** - S_OK / S_ERROR
            result of the command. It is expected ( iff S_OK ) a dictionary like
            { 'Total' : .., 'Free' : ..}

        :return:
          {
            'Status':Error|Active|Bad|Banned,
            'Reason': Some lame statements that have to be updated
          }
        """

        result = {}

        if not commandResult["OK"]:
            result["Status"] = "Error"
            result["Reason"] = commandResult["Message"]
            return S_OK(result)

        commandResult = commandResult["Value"]

        if not commandResult:
            result["Status"] = "Unknown"
            result["Reason"] = "No values to take a decision"
            return S_OK(result)

        for key in ["Total", "Free"]:
            if key not in commandResult:
                result["Status"] = "Error"
                result["Reason"] = "Key %s missing" % key
                return S_OK(result)

        free = float(commandResult["Free"])

        # Units (TB, GB, MB) may change,
        # depending on the configuration of the command in Configurations.py
        if free < 0.1:
            result["Status"] = "Banned"
            result["Reason"] = "Too little free space"
        elif free < 5:
            result["Status"] = "Degraded"
            result["Reason"] = "Little free space"
        else:
            result["Status"] = "Active"
            result["Reason"] = "Enough free space"

        return S_OK(result)



########################################################################
# File: ReplicateAndRegister.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/03/13 18:49:12
########################################################################
""" :mod: ReplicateAndRegister

    ==========================

    .. module: ReplicateAndRegister

    :synopsis: ReplicateAndRegister operation handler

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    ReplicateAndRegister operation handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"
# #
# @file ReplicateAndRegister.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/13 18:49:28
# @brief Definition of ReplicateAndRegister class.

# # imports
import re
from collections import defaultdict

# # from DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Adler import compareAdler, hexAdlerToInt, intAdlerToHex
from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor

from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.DataManagementSystem.Agent.RequestOperations.DMSRequestOperationsBase import DMSRequestOperationsBase

from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

from DIRAC.DataManagementSystem.Client.FTS3Operation import FTS3TransferOperation
from DIRAC.DataManagementSystem.Client.FTS3File import FTS3File
from DIRAC.DataManagementSystem.Client.FTS3Client import FTS3Client
from DIRAC.DataManagementSystem.private.FTS3Utilities import getFTS3Plugin

from DIRAC.ConfigurationSystem.Client.Helpers import Registry

from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter


def filterReplicas(opFile, logger=None, dataManager=None):
    """filter out banned/invalid source SEs"""

    if logger is None:
        logger = gLogger
    if dataManager is None:
        dataManager = DataManager()

    log = logger.getSubLogger("filterReplicas")
    result = defaultdict(list)

    replicas = dataManager.getActiveReplicas(opFile.LFN, getUrl=False, preferDisk=True)
    if not replicas["OK"]:
        log.error("Failed to get active replicas", replicas["Message"])
        return replicas
    reNotExists = re.compile(r".*such file.*")
    replicas = replicas["Value"]
    failed = replicas["Failed"].get(opFile.LFN, "")
    if reNotExists.match(failed.lower()):
        opFile.Status = "Failed"
        opFile.Error = failed
        return S_ERROR(failed)

    replicas = replicas["Successful"].get(opFile.LFN, {})
    noReplicas = False
    if not replicas:
        allReplicas = dataManager.getReplicas(opFile.LFN, getUrl=False)
        if allReplicas["OK"]:
            allReplicas = allReplicas["Value"]["Successful"].get(opFile.LFN, {})
            if not allReplicas:
                result["NoReplicas"].append(None)
                noReplicas = True
            else:
                # There are replicas but we cannot get metadata because the replica is not active
                result["NoActiveReplicas"] += list(allReplicas)
            log.verbose("File has no%s replica in File Catalog" % ("" if noReplicas else " active"), opFile.LFN)
        else:
            return allReplicas

    if not opFile.Checksum or hexAdlerToInt(opFile.Checksum) is False:
        # Set Checksum to FC checksum if not set in the request
        fcMetadata = FileCatalog().getFileMetadata(opFile.LFN)
        fcChecksum = fcMetadata.get("Value", {}).get("Successful", {}).get(opFile.LFN, {}).get("Checksum")
        # Replace opFile.Checksum if it doesn't match a valid FC checksum
        if fcChecksum:
            if hexAdlerToInt(fcChecksum) is not False:
                opFile.Checksum = fcChecksum
                opFile.ChecksumType = fcMetadata["Value"]["Successful"][opFile.LFN].get("ChecksumType", "Adler32")
            else:
                opFile.Checksum = None

    # If no replica was found, return what we collected as information
    if not replicas:
        return S_OK(result)

    for repSEName in replicas:
        repSEMetadata = StorageElement(repSEName).getFileMetadata(opFile.LFN)
        error = repSEMetadata.get("Message", repSEMetadata.get("Value", {}).get("Failed", {}).get(opFile.LFN))
        if error:
            log.warn("unable to get metadata at %s for %s" % (repSEName, opFile.LFN), error.replace("\n", ""))
            if "File does not exist" in error or "No such file" in error:
                result["NoReplicas"].append(repSEName)
            else:
                result["NoMetadata"].append(repSEName)
        elif not noReplicas:
            repSEMetadata = repSEMetadata["Value"]["Successful"][opFile.LFN]

            seChecksum = hexAdlerToInt(repSEMetadata.get("Checksum"))
            # As from here seChecksum is an integer or False, not a hex string!
            if seChecksum is False and opFile.Checksum:
                result["NoMetadata"].append(repSEName)
            elif not seChecksum and opFile.Checksum:
                opFile.Checksum = None
                opFile.ChecksumType = None
            elif seChecksum and (not opFile.Checksum or opFile.Checksum == "False"):
                # Use the SE checksum (convert to hex) and force type to be Adler32
                opFile.Checksum = intAdlerToHex(seChecksum)
                opFile.ChecksumType = "Adler32"
            if not opFile.Checksum or not seChecksum or compareAdler(intAdlerToHex(seChecksum), opFile.Checksum):
                # # All checksums are OK
                result["Valid"].append(repSEName)
            else:
                log.warn(
                    " %s checksum mismatch, FC: '%s' @%s: '%s'"
                    % (opFile.LFN, opFile.Checksum, repSEName, intAdlerToHex(seChecksum))
                )
                result["Bad"].append(repSEName)
        else:
            # If a replica was found somewhere, don't set the file as no replicas
            result["NoReplicas"] = []

    return S_OK(result)


########################################################################
class ReplicateAndRegister(DMSRequestOperationsBase):
    """
    .. class:: ReplicateAndRegister

    ReplicateAndRegister operation handler
    """

    def __init__(self, operation=None, csPath=None):
        """c'tor

        :param self: self reference
        :param Operation operation: Operation instance
        :param str csPath: CS path for this handler
        """
        super(ReplicateAndRegister, self).__init__(operation, csPath)

        # # SE cache

        # Clients
        self.fc = FileCatalog()

    def __call__(self):
        """call me maybe"""

        # The flag  'rmsMonitoring' is set by the RequestTask and is False by default.
        # Here we use 'createRMSRecord' to create the ES record which is defined inside OperationHandlerBase.
        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
        else:
            # # own gMonitor stuff for files
            gMonitor.registerActivity(
                "ReplicateAndRegisterAtt",
                "Replicate and register attempted",
                "RequestExecutingAgent",
                "Files/min",
                gMonitor.OP_SUM,
            )
            gMonitor.registerActivity(
                "ReplicateOK", "Replications successful", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "ReplicateFail", "Replications failed", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RegisterOK", "Registrations successful", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RegisterFail", "Registrations failed", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            # # for FTS
            gMonitor.registerActivity(
                "FTSScheduleAtt", "Files schedule attempted", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "FTSScheduleOK", "File schedule successful", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "FTSScheduleFail", "File schedule failed", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )

        # # check replicas first
        checkReplicas = self.__checkReplicas()
        if not checkReplicas["OK"]:
            self.log.error("Failed to check replicas", checkReplicas["Message"])
        if hasattr(self, "FTSMode") and getattr(self, "FTSMode"):
            bannedGroups = getattr(self, "FTSBannedGroups") if hasattr(self, "FTSBannedGroups") else ()
            if self.request.OwnerGroup in bannedGroups:
                self.log.verbose("usage of FTS system is banned for request's owner")
                return self.dmTransfer()

            return self.fts3Transfer()

        return self.dmTransfer()

    def __checkReplicas(self):
        """check done replicas and update file states"""
        waitingFiles = dict(
            [(opFile.LFN, opFile) for opFile in self.operation if opFile.Status in ("Waiting", "Scheduled")]
        )
        targetSESet = set(self.operation.targetSEList)

        replicas = self.fc.getReplicas(list(waitingFiles))
        if not replicas["OK"]:
            self.log.error("Failed to get replicas", replicas["Message"])
            return replicas

        reMissing = re.compile(r".*such file.*")
        for failedLFN, errStr in replicas["Value"]["Failed"].items():
            waitingFiles[failedLFN].Error = errStr
            if reMissing.search(errStr.lower()):
                self.log.error("File does not exists", failedLFN)
                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                else:
                    gMonitor.addMark("ReplicateFail", len(targetSESet))
                waitingFiles[failedLFN].Status = "Failed"

        for successfulLFN, reps in replicas["Value"]["Successful"].items():
            if targetSESet.issubset(set(reps)):
                self.log.info("file replicated to all targets", successfulLFN)
                waitingFiles[successfulLFN].Status = "Done"

        return S_OK()

    def _addMetadataToFiles(self, toSchedule):
        """Add metadata to those files that need to be scheduled through FTS

        toSchedule is a dictionary:
        {'lfn1': opFile, 'lfn2': opFile}
        """
        if toSchedule:
            self.log.info("found %s files to schedule, getting metadata from FC" % len(toSchedule))
        else:
            self.log.verbose("No files to schedule")
            return S_OK([])

        res = self.fc.getFileMetadata(list(toSchedule))
        if not res["OK"]:
            return res
        else:
            if res["Value"]["Failed"]:
                self.log.warn(
                    "Can't schedule %d files: problems getting the metadata: %s"
                    % (len(res["Value"]["Failed"]), ", ".join(res["Value"]["Failed"]))
                )
            metadata = res["Value"]["Successful"]

        filesToSchedule = {}

        for lfn, lfnMetadata in metadata.items():
            opFileToSchedule = toSchedule[lfn][0]
            opFileToSchedule.GUID = lfnMetadata["GUID"]
            # In principle this is defined already in filterReplicas()
            if not opFileToSchedule.Checksum:
                opFileToSchedule.Checksum = metadata[lfn]["Checksum"]
                opFileToSchedule.ChecksumType = metadata[lfn]["ChecksumType"]
            opFileToSchedule.Size = metadata[lfn]["Size"]

            filesToSchedule[opFileToSchedule.LFN] = opFileToSchedule

        return S_OK(filesToSchedule)

    def _filterReplicas(self, opFile):
        """filter out banned/invalid source SEs"""
        return filterReplicas(opFile, logger=self.log, dataManager=self.dm)

    def _checkExistingFTS3Operations(self):
        """
        Check if there are ongoing FTS3Operation for the current RMS Operation

        Under some conditions, we can be trying to schedule files while
        there is still an FTS transfer going on. This typically happens
        when the REA hangs. To prevent further race condition, we check
        if there are FTS3Operations in a non Final state matching the
        current operation ID. If so, we put the corresponding files in
        scheduled mode. We will then wait till the FTS3 Operation performs
        the callback

        :returns: S_OK with True if we can go on, False if we should stop the processing
        """

        res = FTS3Client().getOperationsFromRMSOpID(self.operation.OperationID)

        if not res["OK"]:
            self.log.debug("Could not get FTS3Operations matching OperationID", self.operation.OperationID)
            return res

        existingFTSOperations = res["Value"]
        # It is ok to have FTS Operations in a final state, so we
        # care only about the others
        unfinishedFTSOperations = [
            ops for ops in existingFTSOperations if ops.status not in FTS3TransferOperation.FINAL_STATES
        ]

        if not unfinishedFTSOperations:
            self.log.debug("No ongoing FTS3Operations, all good")
            return S_OK(True)

        self.log.warn(
            "Some FTS3Operations already exist for the RMS Operation:",
            [op.operationID for op in unfinishedFTSOperations],
        )

        # This would really be a screwed up situation !
        if len(unfinishedFTSOperations) > 1:
            self.log.warn("That's a serious problem !!")

        # We take the rmsFileID of the files in the Operations,
        # find the corresponding File object, and set them scheduled
        rmsFileIDsToSetScheduled = set(
            [ftsFile.rmsFileID for ftsOp in unfinishedFTSOperations for ftsFile in ftsOp.ftsFiles]
        )

        for opFile in self.operation:
            # If it is in the DB, it has a FileID
            opFileID = opFile.FileID
            if opFileID in rmsFileIDsToSetScheduled:
                self.log.warn("Setting RMSFile as already scheduled", opFileID)
                opFile.Status = "Scheduled"

        # We return here such that the Request is set back to Scheduled in the DB
        # With no further modification
        return S_OK(False)

    def fts3Transfer(self):
        """replicate and register using FTS3"""

        self.log.info("scheduling files in FTS3...")

        # Check first if we do not have ongoing transfers

        res = self._checkExistingFTS3Operations()
        if not res["OK"]:
            return res

        # if res['Value'] is False
        # it means that there are ongoing transfers
        # and we should stop here
        if res["Value"] is False:
            # return S_OK such that the request is put back
            return S_OK()

        fts3Files = []
        toSchedule = {}

        # Dict which maps the FileID to the object
        rmsFilesIds = {}

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", len(self.getWaitingFilesList())))

        for opFile in self.getWaitingFilesList():
            rmsFilesIds[opFile.FileID] = opFile

            opFile.Error = ""
            if not self.rmsMonitoring:
                gMonitor.addMark("FTSScheduleAtt")
            # # check replicas
            replicas = self._filterReplicas(opFile)
            if not replicas["OK"]:
                continue
            replicas = replicas["Value"]

            validReplicas = replicas["Valid"]
            noMetaReplicas = replicas["NoMetadata"]
            noReplicas = replicas["NoReplicas"]
            badReplicas = replicas["Bad"]
            noPFN = replicas["NoPFN"]

            if validReplicas:
                validTargets = list(set(self.operation.targetSEList) - set(validReplicas))
                if not validTargets:
                    self.log.info("file %s is already present at all targets" % opFile.LFN)
                    opFile.Status = "Done"
                else:
                    toSchedule[opFile.LFN] = [opFile, validTargets]

            else:
                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                else:
                    gMonitor.addMark("FTSScheduleFail")
                if noMetaReplicas:
                    self.log.warn(
                        "unable to schedule file",
                        "'%s': couldn't get metadata at %s" % (opFile.LFN, ",".join(noMetaReplicas)),
                    )
                    opFile.Error = "Couldn't get metadata"
                elif noReplicas:
                    self.log.error(
                        "Unable to schedule transfer",
                        "File %s doesn't exist at %s" % (opFile.LFN, ",".join(noReplicas)),
                    )
                    opFile.Error = "No replicas found"
                    opFile.Status = "Failed"
                elif badReplicas:
                    self.log.error(
                        "Unable to schedule transfer",
                        "File %s, all replicas have a bad checksum at %s" % (opFile.LFN, ",".join(badReplicas)),
                    )
                    opFile.Error = "All replicas have a bad checksum"
                    opFile.Status = "Failed"
                elif noPFN:
                    self.log.warn("unable to schedule %s, could not get a PFN at %s" % (opFile.LFN, ",".join(noPFN)))

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        res = self._addMetadataToFiles(toSchedule)
        if not res["OK"]:
            return res
        else:
            filesToSchedule = res["Value"]

            for lfn in filesToSchedule:
                opFile = filesToSchedule[lfn]
                validTargets = toSchedule[lfn][1]
                for targetSE in validTargets:
                    ftsFile = FTS3File.fromRMSFile(opFile, targetSE)
                    fts3Files.append(ftsFile)

        if fts3Files:
            res = Registry.getUsernameForDN(self.request.OwnerDN)
            if not res["OK"]:
                self.log.error("Cannot get username for DN", "%s %s" % (self.request.OwnerDN, res["Message"]))
                return res

            username = res["Value"]
            fts3Operation = FTS3TransferOperation.fromRMSObjects(self.request, self.operation, username)
            fts3Operation.ftsFiles = fts3Files

            try:
                if not fts3Operation.activity:
                    vo = getVOfromProxyGroup().get("Value")
                    fts3Plugin = getFTS3Plugin(vo=vo)
                    fts3Operation.activity = fts3Plugin.inferFTSActivity(fts3Operation, self.request, self.operation)
            except Exception:
                pass

            ftsSchedule = FTS3Client().persistOperation(fts3Operation)
            if not ftsSchedule["OK"]:
                self.log.error("Completely failed to schedule to FTS3:", ftsSchedule["Message"])
                return ftsSchedule

            # might have nothing to schedule
            ftsSchedule = ftsSchedule["Value"]
            self.log.info("Scheduled with FTS3Operation id %s" % ftsSchedule)

            self.log.info("%d files have been scheduled to FTS3" % len(fts3Files))

            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", len(fts3Files)))

            for ftsFile in fts3Files:
                opFile = rmsFilesIds[ftsFile.rmsFileID]
                if not self.rmsMonitoring:
                    gMonitor.addMark("FTSScheduleOK", 1)
                opFile.Status = "Scheduled"
                self.log.debug("%s has been scheduled for FTS" % opFile.LFN)
        else:
            self.log.info("No files to schedule after metadata checks")

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        # Just in case some transfers could not be scheduled, try them with RM
        return self.dmTransfer(fromFTS=True)

    def dmTransfer(self, fromFTS=False):
        """replicate and register using dataManager"""
        # # get waiting files. If none just return
        # # source SE
        sourceSE = self.operation.SourceSE if self.operation.SourceSE else None
        if sourceSE:
            # # check source se for read
            bannedSource = self.checkSEsRSS(sourceSE, "ReadAccess")
            if not bannedSource["OK"]:
                if self.rmsMonitoring:
                    for status in ["Attempted", "Failed"]:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(self.operation)))
                    self.rmsMonitoringReporter.commit()
                else:
                    gMonitor.addMark("ReplicateAndRegisterAtt", len(self.operation))
                    gMonitor.addMark("ReplicateFail", len(self.operation))
                return bannedSource

            if bannedSource["Value"]:
                self.operation.Error = "SourceSE %s is banned for reading" % sourceSE
                self.log.info(self.operation.Error)
                return S_OK(self.operation.Error)

        # # check targetSEs for write
        bannedTargets = self.checkSEsRSS()
        if not bannedTargets["OK"]:
            if self.rmsMonitoring:
                for status in ["Attempted", "Failed"]:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(self.operation)))
                self.rmsMonitoringReporter.commit()
            else:
                gMonitor.addMark("ReplicateAndRegisterAtt", len(self.operation))
                gMonitor.addMark("ReplicateFail", len(self.operation))
            return bannedTargets

        if bannedTargets["Value"]:
            self.operation.Error = "%s targets are banned for writing" % ",".join(bannedTargets["Value"])
            return S_OK(self.operation.Error)

        # Can continue now
        self.log.verbose("No targets banned for writing")

        waitingFiles = self.getWaitingFilesList()
        if not waitingFiles:
            return S_OK()
        # # loop over files
        if fromFTS:
            self.log.info("Trying transfer using replica manager as FTS failed")
        else:
            self.log.info("Transferring files using Data manager...")
        errors = defaultdict(int)
        delayExecution = 0

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", len(waitingFiles)))

        for opFile in waitingFiles:
            if opFile.Error in (
                "Couldn't get metadata",
                "File doesn't exist",
                "No active replica found",
                "All replicas have a bad checksum",
            ):
                err = "File already in error status"
                errors[err] += 1

            if not self.rmsMonitoring:
                gMonitor.addMark("ReplicateAndRegisterAtt", 1)

            opFile.Error = ""
            lfn = opFile.LFN

            # Check if replica is at the specified source
            replicas = self._filterReplicas(opFile)
            if not replicas["OK"]:
                self.log.error("Failed to check replicas", replicas["Message"])
                continue
            replicas = replicas["Value"]
            validReplicas = replicas.get("Valid")
            noMetaReplicas = replicas.get("NoMetadata")
            noReplicas = replicas.get("NoReplicas")
            badReplicas = replicas.get("Bad")
            noActiveReplicas = replicas.get("NoActiveReplicas")

            if not validReplicas:
                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                else:
                    gMonitor.addMark("ReplicateFail")
                if noMetaReplicas:
                    err = "Couldn't get metadata"
                    errors[err] += 1
                    self.log.verbose(
                        "unable to replicate '%s', couldn't get metadata at %s" % (opFile.LFN, ",".join(noMetaReplicas))
                    )
                    opFile.Error = err
                elif noReplicas:
                    err = "File doesn't exist"
                    errors[err] += 1
                    self.log.verbose(
                        "Unable to replicate", "File %s doesn't exist at %s" % (opFile.LFN, ",".join(noReplicas))
                    )
                    opFile.Error = err
                    opFile.Status = "Failed"
                elif badReplicas:
                    err = "All replicas have a bad checksum"
                    errors[err] += 1
                    self.log.error(
                        "Unable to replicate",
                        "%s, all replicas have a bad checksum at %s" % (opFile.LFN, ",".join(badReplicas)),
                    )
                    opFile.Error = err
                    opFile.Status = "Failed"
                elif noActiveReplicas:
                    err = "No active replica found"
                    errors[err] += 1
                    self.log.verbose(
                        "Unable to schedule transfer", "%s, %s at %s" % (opFile.LFN, err, ",".join(noActiveReplicas))
                    )
                    opFile.Error = err
                    # All source SEs are banned, delay execution by 1 hour
                    delayExecution = 60
                continue
            # # get the first one in the list
            if sourceSE not in validReplicas:
                if sourceSE:
                    err = "File not at specified source"
                    errors[err] += 1
                    self.log.warn(
                        "%s is not at specified sourceSE %s, changed to %s" % (lfn, sourceSE, validReplicas[0])
                    )
                sourceSE = validReplicas[0]

            # # loop over targetSE
            catalogs = self.operation.Catalog
            if catalogs:
                catalogs = [cat.strip() for cat in catalogs.split(",")]

            for targetSE in self.operation.targetSEList:

                # # call DataManager
                if targetSE in validReplicas:
                    self.log.warn("Request to replicate %s to an existing location: %s" % (lfn, targetSE))
                    continue
                res = self.dm.replicateAndRegister(lfn, targetSE, sourceSE=sourceSE, catalog=catalogs)
                if res["OK"]:

                    if lfn in res["Value"]["Successful"]:

                        if "replicate" in res["Value"]["Successful"][lfn]:

                            repTime = res["Value"]["Successful"][lfn]["replicate"]
                            prString = "file %s replicated at %s in %s s." % (lfn, targetSE, repTime)

                            if not self.rmsMonitoring:
                                gMonitor.addMark("ReplicateOK", 1)

                            if "register" in res["Value"]["Successful"][lfn]:

                                if not self.rmsMonitoring:
                                    gMonitor.addMark("RegisterOK", 1)

                                regTime = res["Value"]["Successful"][lfn]["register"]
                                prString += " and registered in %s s." % regTime
                                self.log.info(prString)
                            else:

                                if not self.rmsMonitoring:
                                    gMonitor.addMark("RegisterFail", 1)
                                prString += " but failed to register"
                                self.log.warn(prString)

                                opFile.Error = "Failed to register"
                                # # add register replica operation
                                registerOperation = self.getRegisterOperation(opFile, targetSE, type="RegisterReplica")
                                self.request.insertAfter(registerOperation, self.operation)

                        else:

                            self.log.error("Failed to replicate", "%s to %s" % (lfn, targetSE))
                            if not self.rmsMonitoring:
                                gMonitor.addMark("ReplicateFail", 1)
                            opFile.Error = "Failed to replicate"

                    else:

                        if not self.rmsMonitoring:
                            gMonitor.addMark("ReplicateFail", 1)
                        reason = res["Value"]["Failed"][lfn]
                        self.log.error("Failed to replicate and register", "File %s at %s:" % (lfn, targetSE), reason)
                        opFile.Error = reason

                else:

                    if not self.rmsMonitoring:
                        gMonitor.addMark("ReplicateFail", 1)
                    opFile.Error = "DataManager error: %s" % res["Message"]
                    self.log.error("DataManager error", res["Message"])

            if not opFile.Error:
                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", 1))

                if len(self.operation.targetSEList) > 1:
                    self.log.info("file %s has been replicated to all targetSEs" % lfn)
                opFile.Status = "Done"
            elif self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
        # Log error counts
        if delayExecution:
            self.log.info("Delay execution of the request by %d minutes" % delayExecution)
            self.request.delayNextExecution(delayExecution)
        for error, count in errors.items():
            self.log.error(error, "for %d files" % count)

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        return S_OK()



""" Tornado-based HTTPs JobMonitoring service.
"""


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger
from DIRAC.Core.Tornado.Server.TornadoService import TornadoService
from DIRAC.WorkloadManagementSystem.Service.JobMonitoringHandler import JobMonitoringHandlerMixin


sLog = gLogger.getSubLogger(__name__)


class TornadoJobMonitoringHandler(JobMonitoringHandlerMixin, TornadoService):
    log = sLog



#!/usr/bin/env python
"""
Determine Normalization for current CPU. The main users of this script are the pilot jobs.

Pilots invoke dirac-wms-cpu-normalization which
- runs 1 iteration of singleDiracBenchmark(1) (for single processors only)
- stores in local cfg the following::

    LocalSite
    {
      CPUNormalizationFactor = 23.7 # corrected value (by JobScheduling/CPUNormalizationCorrection)
      DB12measured = 15.4
    }

DB12measured is up to now wrote down but never used.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

from DIRAC import gLogger, gConfig
from DIRAC.WorkloadManagementSystem.Client.DIRACbenchmark import singleDiracBenchmark
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations


@Script()
def main():
    Script.registerSwitch("U", "Update", "Update dirac.cfg with the resulting value")
    Script.registerSwitch("R:", "Reconfig=", "Update given configuration file with the resulting value")
    Script.parseCommandLine(ignoreErrors=True)

    update = False
    configFile = None

    for unprocSw in Script.getUnprocessedSwitches():
        if unprocSw[0] in ("U", "Update"):
            update = True
        elif unprocSw[0] in ("R", "Reconfig"):
            configFile = unprocSw[1]

    result = singleDiracBenchmark(1)

    if result is None:
        gLogger.error("Cannot make benchmark measurements")
        DIRAC.exit(1)

    db12Measured = round(result["NORM"], 1)
    corr = Operations().getValue("JobScheduling/CPUNormalizationCorrection", 1.0)
    norm = round(result["NORM"] / corr, 1)

    gLogger.notice("Estimated CPU power is %.1f HS06" % norm)

    if update:
        gConfig.setOptionValue("/LocalSite/CPUNormalizationFactor", norm)
        gConfig.setOptionValue("/LocalSite/DB12measured", db12Measured)

        if configFile:
            gConfig.dumpLocalCFGToFile(configFile)
        else:
            gConfig.dumpLocalCFGToFile(gConfig.diracConfigFilePath)

    DIRAC.exit()


if __name__ == "__main__":
    main()



""" This module exposes singleton gServiceInterface as istance of ServiceInterface (also in this module)

    Interacts with RRD (rrdtool), with ComponentMonitoringDB (mysql) and with MonitoringCatalog (sqlite3)

    Main clients are the monitoring handler (what's called by gMonitor object), and the web portal.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from past.builtins import long
from DIRAC import S_OK, S_ERROR

from DIRAC import gLogger, rootPath, gConfig
from DIRAC.Core.Utilities import DEncode, List

from DIRAC.FrameworkSystem.private.monitoring.RRDManager import RRDManager
from DIRAC.FrameworkSystem.private.monitoring.PlotCache import PlotCache
from DIRAC.FrameworkSystem.DB.ComponentMonitoringDB import ComponentMonitoringDB
from DIRAC.FrameworkSystem.private.monitoring.MonitoringCatalog import MonitoringCatalog


class ServiceInterface(object):
    """
    The gServiceInterface singleton object which is used by the MonitoringHandle to register activities and
    commit the data is inherited from this class.
    """

    __sourceToComponentIdMapping = {}

    def __init__(self):
        self.dataPath = "%s/data/monitoring" % gConfig.getValue("/LocalSite/InstancePath", rootPath)
        self.plotsPath = "%s/plots" % self.dataPath
        self.rrdPath = "%s/rrd" % self.dataPath
        self.srvUp = False
        self.compmonDB = False

    def __createRRDManager(self):
        """
        Generates an RRDManager
        """
        return RRDManager(self.rrdPath, self.plotsPath)

    def __createCatalog(self):
        """
        Creates a Monitoring catalog connector
        """
        return MonitoringCatalog(self.dataPath)

    def initialize(self, dataPath):
        """
        Initializes monitoring server

        :type dataPath: string
        :param dataPath: The root path of the data directory.
        """

        self.dataPath = dataPath
        self.plotCache = PlotCache(RRDManager(self.rrdPath, self.plotsPath))
        self.srvUp = True
        try:
            self.compmonDB = ComponentMonitoringDB()
        except Exception as e:
            gLogger.exception("Cannot initialize component monitoring db: %s" % e)

    def initializeDB(self):
        """
        Initializes and creates monitoring catalog db if it doesn't exist
        """
        acCatalog = self.__createCatalog()
        if not acCatalog.createSchema():
            return False
        # Register default view if it's not there
        viewNames = [str(v[1]) for v in acCatalog.getViews(False)]
        if "Dynamic component view" in viewNames:
            return True
        return acCatalog.registerView(
            "Dynamic component view",
            DEncode.encode(
                {
                    "variable": ["sources.componentName"],
                    "definition": {},
                    "stacked": True,
                    "groupBy": ["activities.description"],
                    "label": "$SITE",
                }
            ),
            ["sources.componentName"],
        )

    def __checkSourceDict(self, sourceDict):
        """
        Checks that the dictionary is a valid source one

        :type sourceDict: dictionary
        :param sourceDict: The source dictionary containing information about "setup", "site", "componentType",
                                                                              "componentLocation", "componentName"
        :return: bool
        """
        validKeys = ("setup", "site", "componentType", "componentLocation", "componentName")
        for key in validKeys:
            if key not in sourceDict:
                return False
        return True

    def __checkActivityDict(self, acDict):
        """
        Checks that the dictionary is a valid activity one

        :type acDict: dictionary
        :param acDict: The activity dictionary containing information about 'category', 'description', 'bucketLength',
                                                                            'type', 'unit'.
        :return: bool
        """
        validKeys = ("category", "description", "bucketLength", "type", "unit")
        for key in validKeys:
            if key not in acDict:
                return False
        return True

    def registerActivities(self, sourceDict, activitiesDict, componentExtraInfo):
        """
        Registers new activities in the database.

        :type sourceDict: dictionary
        :param sourceDict: The source dictionary containing information about "setup", "site", "componentType",
                                                                              "componentLocation", "componentName"
        :type activitiesDict: dictionary
        :param activitiesDict: The activity dictionary containing information about 'category', 'description',
                                                                                    'bucketLength', 'type', 'unit'
        :type componentExtraInfo: dictionary
        :param componentExtraInfo: This contains one or many of these  'version', 'DIRACVersion', 'description',
                                                                       'startTime', 'platform', 'cycles', 'queries'.
        :return: S_OK with sourceId / S_ERROR with the corresoponding message.
        """
        acCatalog = self.__createCatalog()
        rrdManager = self.__createRRDManager()
        # Register source
        if not self.__checkSourceDict(sourceDict):
            return S_ERROR("Source definition is not valid")
        sourceId = acCatalog.registerSource(sourceDict)
        # Register activities
        for name in activitiesDict:
            if not self.__checkActivityDict(activitiesDict[name]):
                return S_ERROR("Definition for activity %s is not valid" % name)
            activitiesDict[name]["name"] = name
            if "bucketLength" not in activitiesDict[name]:
                activitiesDict[name]["bucketLength"] = 60
            if not self.__checkActivityDict(activitiesDict[name]):
                return S_ERROR("Activity %s definition is not valid" % name)
            gLogger.info("Received activity", "%s [%s]" % (name, str(activitiesDict[name])))
            rrdFile = acCatalog.registerActivity(sourceId, name, activitiesDict[name])
            if not rrdFile:
                return S_ERROR("Could not register activity %s" % name)
            retVal = rrdManager.create(activitiesDict[name]["type"], rrdFile, activitiesDict[name]["bucketLength"])
            if not retVal["OK"]:
                return retVal
        self.__cmdb_registerComponent(sourceId, sourceDict, componentExtraInfo)
        return S_OK(sourceId)

    def commitMarks(self, sourceId, activitiesDict, componentExtraInfo):
        """
        Adds marks to activities.

        :type sourceId: int
        :param sourceId: This id is returned after the activity is registered inside the database and
                         is used for committing the marks.
        :type activitiesDict: dictionary
        :param activitiesDict: The activity dictionary containing information about 'category', 'description',
                                                                                    'bucketLength', 'type', 'unit'
        :type componentExtraInfo: dictionary
        :param componentExtraInfo: This contains one or many of these  'version', 'DIRACVersion', 'description',
                                                                       'startTime', 'platform', 'cycles', 'queries'.
        :return: S_OK with a list of unregistered activities.
        """
        gLogger.info("Committing marks", "From %s for %s" % (sourceId, ", ".join(activitiesDict.keys())))
        acCatalog = self.__createCatalog()
        rrdManager = self.__createRRDManager()
        unregisteredActivities = []
        for acName in activitiesDict:
            acData = activitiesDict[acName]
            acInfo = acCatalog.findActivity(sourceId, acName)
            if not acInfo:
                unregisteredActivities.append(acName)
                gLogger.warn("Cant find rrd filename", "%s:%s activity" % (sourceId, acName))
                continue
            rrdFile = acInfo[6]
            if not rrdManager.existsRRDFile(rrdFile):
                gLogger.error("RRD file does not exist", "%s:%s activity (%s)" % (sourceId, acName, rrdFile))
                unregisteredActivities.append(acName)
                continue
            gLogger.info("Updating activity", "%s -> %s" % (acName, rrdFile))
            timeList = sorted(acData)
            entries = []
            for instant in timeList:
                entries.append((instant, acData[instant]))
            if len(entries) > 0:
                gLogger.verbose("There are %s entries for %s" % (len(entries), acName))
                retDict = rrdManager.update(acInfo[4], rrdFile, acInfo[7], entries, long(acInfo[8]))
                if not retDict["OK"]:
                    gLogger.error("There was an error updating", "%s:%s activity [%s]" % (sourceId, acName, rrdFile))
                else:
                    acCatalog.setLastUpdate(sourceId, acName, retDict["Value"])
        if not self.__cmdb_heartbeatComponent(sourceId, componentExtraInfo):
            for acName in activitiesDict:
                if acName not in unregisteredActivities:
                    unregisteredActivities.append(acName)
        return S_OK(unregisteredActivities)

    def fieldValue(self, field, definedFields):
        """
        Calculated values for a field given a set of defined values for other fields.

        :type field: list
        :param field: A set of fields.
        :type definedFields: list
        :param definedFields: A set of defined fields
        :return: S_OK with a list of the desired values.
        """
        retList = self.__createCatalog().queryField("DISTINCT %s" % field, definedFields)
        return S_OK(retList)

    def __getGroupedPlots(self, viewDescription):
        """
        Calculates grouped plots for a view.

        :type viewDescription: dictionary
        :param viewDescription: A dictionary containing the view description.
        :return: list of plots.
        """
        plotsList = []
        acCatalog = self.__createCatalog()
        groupList = acCatalog.queryField(
            "DISTINCT %s" % ", ".join(viewDescription["groupBy"]), viewDescription["definition"]
        )
        for grouping in groupList:
            gLogger.debug("Grouped plot for combination %s" % str(grouping))
            groupDefinitionDict = dict(viewDescription["definition"])
            for index in range(len(viewDescription["groupBy"])):
                groupDefinitionDict[viewDescription["groupBy"][index]] = grouping[index]
            activityList = acCatalog.getMatchingActivities(groupDefinitionDict)
            for activity in activityList:
                activity.setGroup(grouping)
                activity.setLabel(viewDescription["label"])
            plotsList.append(activityList)
        return plotsList

    def __generateGroupPlots(self, fromSecs, toSecs, viewDescription, size):
        """
        Generates grouped plots for a view.

        :type fromSecs: int
        :param fromSecs: A value in seconds from where to start.
        :type toSecs: int
        :param toSecs: A value in seconds for where to end.
        :type viewDescription: dictionary
        :param viewDescription: A dictionary containing the view description.
        :type size: int
        :param size: There is a matrix defined for size so here only one of these values go [0, 1, 2, 3].
        :return: S_OK with a list of files.
        """
        plotList = self.__getGroupedPlots(viewDescription)
        filesList = []
        for groupPlot in plotList:
            retVal = self.plotCache.groupPlot(fromSecs, toSecs, groupPlot, viewDescription["stacked"], size)
            if not retVal["OK"]:
                gLogger.error("There was a problem ploting", retVal["Message"])
                return retVal
            graphFile = retVal["Value"]
            gLogger.verbose("Generated graph", "file %s for group %s" % (graphFile, str(groupPlot[0])))
            filesList.append(graphFile)
        return S_OK(filesList)

    def __getPlots(self, viewDescription):
        """
        Calculates plots for a view.

        :type viewDescription: dictionary
        :param viewDescription: A dictionary containing the view description.
        :return: list of matching activities.
        """
        acCatalog = self.__createCatalog()
        return acCatalog.getMatchingActivities(viewDescription["definition"])

    def __generatePlots(self, fromSecs, toSecs, viewDescription, size):
        """
        Generates non grouped plots for a view.

        :type fromSecs: int
        :param fromSecs: A value in seconds from where to start.
        :type toSecs: int
        :param toSecs: A value in seconds for where to end.
        :type viewDescription: dictionary
        :param viewDescription: A dictionary containing the view description.
        :type size: int
        :param size: There is a matrix defined for size so here only one of these values go [0, 1, 2, 3].
        :return: S_OK with a list of files.
        """
        acList = self.__getPlots(viewDescription)
        filesList = []
        for activity in acList:
            activity.setLabel(viewDescription["label"])
            retVal = self.plotCache.plot(fromSecs, toSecs, activity, viewDescription["stacked"], size)
            if not retVal["OK"]:
                gLogger.error("There was a problem ploting", retVal["Message"])
                return retVal
            graphFile = retVal["Value"]
            gLogger.verbose("Generated graph", "file %s" % (graphFile))
            filesList.append(graphFile)
        return S_OK(filesList)

    def generatePlots(self, fromSecs, toSecs, viewDescription, size=1):
        """
        Generates plots for a view.

        :type fromSecs: int
        :param fromSecs: A value in seconds from where to start.
        :type toSecs: int
        :param toSecs: A value in seconds for where to end.
        :type viewDescription: dictionary
        :param viewDescription: A dictionary containing the view description.
        :type size: int
        :param size: There is a matrix defined for size so here only one of these values go [0, 1, 2, 3].
        :return: S_OK with a list of files.
        """
        gLogger.info("Generating plots", str(viewDescription))
        if "stacked" not in viewDescription:
            viewDescription["stacked"] = False
        if "label" not in viewDescription:
            viewDescription["label"] = ""
        if "groupBy" in viewDescription and len(viewDescription["groupBy"]):
            return self.__generateGroupPlots(fromSecs, toSecs, viewDescription, size)
        return self.__generatePlots(fromSecs, toSecs, viewDescription, size)

    def getGraphData(self, filename):
        """
        Reads the contents of a plot file.

        :type filename: string
        :param filname: The name of the file.
        :return: S_OK with all the data that is read.
        """
        try:
            with open("%s/%s" % (self.plotsPath, filename)) as fd:
                data = fd.read()
        except Exception as e:
            return S_ERROR(e)
        else:
            return S_OK(data)

    def saveView(self, viewName, viewDescription):
        """
        Saves a view in the catalog.

        :type viewName: string
        :param viewName: Name of the view.
        :type viewDescription: dictionary
        :param viewDescription: A dictionary containing the view description.
        :return: S_OK / S_ERROR with the corresponding error message.
        """
        if "stacked" not in viewDescription:
            viewDescription["stacked"] = False
        if "label" not in viewDescription:
            viewDescription["label"] = ""
        if "variable" in viewDescription:
            for varField in viewDescription["variable"]:
                if varField in viewDescription["definition"]:
                    del viewDescription["definition"][varField]
        else:
            viewDescription["variable"] = []
        acCatalog = self.__createCatalog()
        return acCatalog.registerView(viewName, DEncode.encode(viewDescription), viewDescription["variable"])

    def getViews(self, onlyStatic=True):
        """
        Gets all stored views.

        :type onlyStatic: bool
        :param onlyStatic: Whether the views required are static or not.
        :return: S_OK with a list of views.
        """
        viewsList = self.__createCatalog().getViews(onlyStatic)
        return S_OK(viewsList)

    def plotView(self, viewRequest):
        """
        Generates all plots for a view.

        :type viewRequest: dictionary
        :param viewRequest: Containing information of the view like 'id', 'fromSecs', 'toSecs', etc.
        :return: S_OK with a list of files.
        """
        views = self.__createCatalog().getViewById(viewRequest["id"])
        if len(views) == 0:
            return S_ERROR("View does not exist")
        viewData = views[0]
        viewDefinition = DEncode.decode(str(viewData[0]))[0]
        neededVarFields = List.fromChar(viewData[1], ",")
        if len(neededVarFields) > 0:
            if "varData" not in viewRequest:
                return S_ERROR("Missing variable fields %s!" % ", ".join(neededVarFields))
            missingVarFields = []
            for neededField in neededVarFields:
                if neededField in viewRequest["varData"]:
                    viewDefinition["definition"][neededField] = viewRequest["varData"][neededField]
                else:
                    missingVarFields.append(neededField)
            if len(missingVarFields) > 0:
                return S_ERROR("Missing required fields %s!" % ", ".join(missingVarFields))
        return self.generatePlots(viewRequest["fromSecs"], viewRequest["toSecs"], viewDefinition, viewRequest["size"])

    def deleteView(self, viewId):
        """
        Deletes a view.

        :type viewId: string
        :param viewId: The view id.
        :return: S_OK
        """
        self.__createCatalog().deleteView(viewId)
        return S_OK()

    def getSources(self, dbCond={}, fields=[]):
        """
        Gets a list of activities.

        :type dbCond: dictionary
        :param dbCond: The required database conditions.
        :type fields: list
        :param fields: A list of required fields.
        :return: The list of results after the query is performed.
        """
        catalog = self.__createCatalog()
        return catalog.getSources(dbCond, fields)

    def getActivities(self, dbCond={}):
        """
        Gets a list of activities.

        :type dbCond: dictionary
        :param dbCond: The required database conditions.
        :return: an activity dictionary.
        """
        acDict = {}
        catalog = self.__createCatalog()
        for sourceTuple in catalog.getSources(dbCond):
            activityCond = {"sourceId": sourceTuple[0]}
            acDict[sourceTuple] = catalog.getActivities(activityCond)
        return acDict

    def getNumberOfActivities(self, dbCond={}):
        """
        Gets the number of activities.

        :type dbCond: dictionary
        :param dbCond: The required database conditions.
        :return: An integer indicating the total number of activities.
        """
        catalog = self.__createCatalog()
        total = 0
        for sourceTuple in catalog.getSources(dbCond):
            activityCond = {"sourceId": sourceTuple[0]}
            total += len(catalog.getActivities(activityCond))
        return total

    def getActivitiesContents(self, selDict, sortList, start, limit):
        """
        Gets all the sources and activities details in a joined format.

        :type selDict: dictionary
        :param selDict: The fields inside the select query.
        :type sortList: list
        :param sortList: A list in sorted order of the data.
        :type start: int
        :param start: The point or tuple from where to start.
        :type limit: int
        :param limit: The number of tuples to select from the starting point.
        :return: S_OK with a tuple of the result list and fields list.
        """
        return self.__createCatalog().activitiesQuery(selDict, sortList, start, limit)

    def deleteActivity(self, sourceId, activityId):
        """
        Deletes a view.

        :type sourceId: string
        :param sourceId: The source id.
        :type activityId: string
        :param activityId: The activity id.
        :return: S_OK / The corresponding error message.
        """
        retVal = self.__createCatalog().deleteActivity(sourceId, activityId)
        if not retVal["OK"]:
            return retVal
        self.__createRRDManager().deleteRRD(retVal["Value"])
        return S_OK()

    # ComponentMonitoringDB functions

    def __cmdb__writeComponent(self, sourceId):
        """
        Used to write basic component details to the ComponentMonitoringDB(MySQLdb).

        :type sourceId: string
        :param sourceId: the source id.
        :return: True / False
        """
        if sourceId not in ServiceInterface.__sourceToComponentIdMapping:
            if not self.__cmdb__loadComponentFromActivityDB(sourceId):
                return False
        compDict = ServiceInterface.__sourceToComponentIdMapping[sourceId]
        result = self.compmonDB.registerComponent(compDict)
        if not result["OK"]:
            gLogger.error("Cannot register component in ComponentMonitoringDB", result["Message"])
            return False
        compDict["compId"] = result["Value"]
        self.__cmdb__writeHeartbeat(sourceId)
        gLogger.info("Registered component in component monitoring db")
        return True

    def __cmdb__merge(self, sourceId, extraDict):
        """
        Merges the cached dict.

        :type sourceId: string
        :param sourceId: the source id.
        :type extraDict: dictionary
        :param extraDict: The dictionary containing extra information.
        """
        compDict = ServiceInterface.__sourceToComponentIdMapping[sourceId]
        for field in self.compmonDB.getOptionalFields():
            if field in extraDict:
                compDict[field] = extraDict[field]
        ServiceInterface.__sourceToComponentIdMapping[sourceId] = compDict

    def __cmdb__loadComponentFromActivityDB(self, sourceId):
        """
        Loads the component dict from the activities it registered.

        :type sourceId: string
        :param sourceId: the source id.
        :return: True / False
        """
        sources = gServiceInterface.getSources(
            {"id": sourceId}, ["componentType", "componentName", "componentLocation", "setup"]
        )
        if len(sources) == 0:
            return False
        source = [ts for ts in sources if len(ts) > 0][0]
        compDict = {
            "type": source[0],
            "componentName": source[1],
            "host": source[2],
            "setup": source[3],
        }
        if compDict["type"] == "service":
            loc = compDict["host"]
            loc = loc[loc.find("://") + 3 :]
            loc = loc[: loc.find("/")]
            compDict["host"] = loc[: loc.find(":")]
            compDict["port"] = loc[loc.find(":") + 1 :]
        ServiceInterface.__sourceToComponentIdMapping[sourceId] = compDict
        return True

    def __cmdb__writeHeartbeat(self, sourceId):
        """
        Writes the heartbeat stamp to the ComponentMonitoringDB.

        :type sourceId: string
        :param sourceId: the source id.
        :return: True / False
        """
        compDict = ServiceInterface.__sourceToComponentIdMapping[sourceId]
        result = self.compmonDB.heartbeat(compDict)
        if not result["OK"]:
            gLogger.error("Cannot heartbeat component in ComponentMonitoringDB", result["Message"])

    def __cmdb_registerComponent(self, sourceId, sourceDict, componentExtraInfo):
        """
        Writes all the basic component information to the ComponentMonitoringDB.

        :type sourceId: string
        :param sourceId: the source id.
        :type sourceDict: dictionary
        :param sourceDict: The dictionary containing source information.
        :type componentExtraInfo: dictionary
        :param componentExtraInfo: The dictionary containing extra information.
        """
        if sourceDict["componentType"] not in ("service", "agent"):
            return
        compDict = {
            "componentName": sourceDict["componentName"],
            "setup": sourceDict["setup"],
            "type": sourceDict["componentType"],
            "host": sourceDict["componentLocation"],
        }
        if compDict["type"] == "service":
            loc = compDict["host"]
            loc = loc[loc.find("://") + 3 :]
            loc = loc[: loc.find("/")]
            compDict["host"] = loc[: loc.find(":")]
            compDict["port"] = loc[loc.find(":") + 1 :]
        ServiceInterface.__sourceToComponentIdMapping[sourceId] = compDict
        self.__cmdb__merge(sourceId, componentExtraInfo)
        self.__cmdb__writeComponent(sourceId)

    def __cmdb_heartbeatComponent(self, sourceId, componentExtraInfo):
        """
        This method is used to write the Component heartbeat using the __cmdb__writeHeartbeat method.

        :type sourceId: string
        :param sourceId: the source id.
        :type componentExtraInfo: dictionary
        :param componentExtraInfo: The dictionary containing extra information.
        :return: True / False
        """
        # Component heartbeat
        if sourceId not in ServiceInterface.__sourceToComponentIdMapping:
            if not self.__cmdb__loadComponentFromActivityDB(sourceId):
                return False
        if ServiceInterface.__sourceToComponentIdMapping[sourceId]["type"] not in ("service", "agent"):
            return True
        self.__cmdb__merge(sourceId, componentExtraInfo)
        self.__cmdb__writeHeartbeat(sourceId)
        return True

    def getComponentsStatus(self, condDict=False):
        """
        This method basically returns the component status by reading it from the ComponentMonitoringDB.

        :type condDict: dictionary
        :param condDict: The dictionary containing the conditions.
        :return: S_OK with the requires results.
        """
        if not condDict:
            condDict = {}
        return self.compmonDB.getComponentsStatus(condDict)


gServiceInterface = ServiceInterface()



"""
   OcciEndpoint is Endpoint base class implementation for the Occi cloud service.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

__RCSID__ = "$Id$"

import os
import requests
from requests.auth import HTTPBasicAuth
import uuid
import base64

# DIRAC
from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Resources.Cloud.Endpoint import Endpoint
from DIRAC.Resources.Cloud.KeystoneClient import KeystoneClient
from DIRAC.Core.Utilities.File import makeGuid

DEBUG = False


class OcciEndpoint(Endpoint):
    """OCCI implementation of the Cloud Endpoint interface"""

    def __init__(self, parameters=None):
        super(OcciEndpoint, self).__init__(parameters=parameters)
        # logger
        self.log = gLogger.getSubLogger("OcciEndpoint")
        self.valid = False
        self.vmType = self.parameters.get("VMType")
        self.site = self.parameters.get("Site")

        # Prepare the authentication request parameters
        self.session = None
        self.authArgs = {}
        self.user = self.parameters.get("User")
        self.password = self.parameters.get("Password")
        self.loginMode = False
        if self.user and self.password:
            # we have the login/password case
            self.authArgs["auth"] = HTTPBasicAuth(self.user, self.password)
            self.authArgs["verify"] = False
            self.loginMode = True
        else:
            # we have the user proxy case
            self.userProxy = os.environ.get("X509_USER_PROXY")
            self.userProxy = self.parameters.get("Proxy", self.userProxy)
            if self.userProxy:
                self.parameters["Proxy"] = self.userProxy
            if self.userProxy is None:
                self.log.error("User proxy is not defined")
                self.valid = False
                return
            self.authArgs["cert"] = self.userProxy
            self.caPath = self.parameters.get("CAPath", "/etc/grid-security/certificates")
            self.parameters["CAPath"] = self.caPath
            self.authArgs["verify"] = self.caPath
            if self.parameters.get("Auth") == "voms":
                self.authArgs["data"] = '{"auth":{"voms": true}}'

        self.serviceUrl = self.parameters.get("EndpointUrl")
        self.computeUrl = None
        self.tenant = self.parameters.get("Tenant")
        self.token = None
        self.scheme = {}
        result = self.initialize()
        if result["OK"]:
            self.log.debug("OcciEndpoint created and validated")
            self.valid = True

        # import pprint
        # pprint.pprint( self.scheme )

    def initialize(self):

        try:
            result = requests.head(self.serviceUrl + "/-/", headers={"Content-Type": "text/plain"}, **self.authArgs)
        except Exception as exc:
            return S_ERROR(repr(exc))

        # for key,value in result.headers.items():
        #  print "AT >>> initialize", key,value
        # print "AT >>> initialize text", result.text

        return self.__checkConnection()

    def __getKeystoneUrl(self):

        # The URL can be specified in the configuration
        if self.parameters.get("KeystoneURL"):
            return S_OK(self.parameters.get("KeystoneURL"))

        # Make a trial service call
        try:
            result = requests.head(self.serviceUrl + "/-/", headers={"Content-Type": "text/plain"}, **self.authArgs)
        except Exception as e:
            return S_ERROR(str(e))

        # print "AT >>> __getKeystoneUrl", result, result.text
        # print "AT >>> __getKeystoneUrl", result.headers

        # This is not an authentication error
        if result.status_code != 401 or result.headers is None:
            return S_OK(None)
            # return S_ERROR('Do not recognise response when connecting to ' + self.serviceUrl)

        if "www-authenticate" not in result.headers:
            return S_OK(None)

        if not result.headers["www-authenticate"].startswith("Keystone uri="):
            return S_ERROR(
                'Only Keystone authentication is currently supported (instead got "%s")'
                % result.headers["www-authenticate"]
            )

        try:
            keystoneURL = result.headers["www-authenticate"][14:-1]
        except BaseException:
            return S_ERROR("Failed to find Keystone URL in %s" % result.headers["www-authenticate"])

        return S_OK(keystoneURL)

    def __getSchemaDefinitions(self):

        try:
            response = self.session.get("%s/-/" % self.serviceUrl, headers={"Accept": "text/plain,text/occi"})

        except Exception as exc:
            return S_ERROR("Failed to get schema definition: %s" % str(exc))

        if response.status_code != 200:
            return S_ERROR("Failed to get schema definition", response.text)

        self.scheme = {}

        categories = response.text.split("\n")[1:]
        for category in categories:
            if category:
                values = category.split(";")
                categoryName = values[0][values[0].find(":") + 1 :].strip()
                scheme = None
                className = None
                title = None
                location = None
                for prop in values:
                    if "scheme=" in prop:
                        scheme = prop.strip().replace("scheme=", "").replace('"', "")
                    if "class=" in prop:
                        className = prop.strip().replace("class=", "").replace('"', "")
                    if "title=" in prop:
                        title = prop.strip().replace("title=", "").replace('"', "")
                    if "location=" in prop:
                        tmp = prop.strip()
                        tmp = tmp.replace("https://", "").replace("http://", "").replace('"', "")
                        tmp = tmp[tmp.find("/") :]
                        location = tmp

                if className is None:
                    return S_ERROR("Failed to get schema definition:", "no class for category %s" % categoryName)
                self.scheme.setdefault(className, {})
                self.scheme[className][categoryName] = {}
                if scheme is not None:
                    self.scheme[className][categoryName]["scheme"] = scheme
                if title is not None:
                    self.scheme[className][categoryName]["title"] = title
                if location is not None:
                    self.scheme[className][categoryName]["location"] = location

        return S_OK()

    def __checkConnection(self):
        """
        Checks connection status by trying to list the images.

        :return: S_OK | S_ERROR
        """
        self.session = requests.Session()
        self.session.mount(self.serviceUrl, requests.adapters.HTTPAdapter(pool_connections=20))
        self.session.verify = self.authArgs["verify"]

        # Retrieve token
        result = self.__getKeystoneUrl()
        if not result["OK"]:
            return result

        self.keystoneUrl = result["Value"]
        if self.keystoneUrl is not None:
            keystoneClient = KeystoneClient(self.keystoneUrl, self.parameters)
            result = keystoneClient.getToken()
            if not result["OK"]:
                return result
            self.token = result["Value"]
            self.session.headers.clear()
            self.session.headers.update({"X-Auth-Token": self.token})
            self.session.verify = self.authArgs["verify"]
        else:
            if self.loginMode:
                self.session.auth = self.authArgs["auth"]
                self.session.verify = self.authArgs["verify"]
            else:
                self.session.cert = self.userProxy
                self.session.verify = self.caPath

        result = self.__getSchemaDefinitions()
        if not result["OK"]:
            return result
        self.computeUrl = "%s/compute/" % (self.serviceUrl)
        return S_OK()

    def createInstances(self, vmsToSubmit):
        outputDict = {}
        message = ""
        for nvm in range(vmsToSubmit):
            instanceID = makeGuid()[:8]
            createPublicIP = "ipPool" in self.parameters
            result = self.createInstance(instanceID, createPublicIP)
            if result["OK"]:
                nodeID, publicIP = result["Value"]
                self.log.debug("Created VM instance %s/%s with publicIP %s" % (nodeID, instanceID, publicIP))
                nodeDict = {}
                nodeDict["PublicIP"] = publicIP
                nodeDict["InstanceID"] = instanceID
                nodeDict["NumberOfProcessors"] = self.parameters.get("NumberOfProcessors", 1)
                # nodeDict['RAM'] = self.flavor.ram
                # nodeDict['DiskSize'] = self.flavor.disk
                # nodeDict['Price'] = self.flavor.price
                outputDict[nodeID] = nodeDict
            else:
                message = result["Message"]
                break

        # We failed submission utterly
        if not outputDict:
            return S_ERROR("No VM submitted: %s" % message)

        return S_OK(outputDict)

    def __renderCategory(self, category, className):

        if className not in self.scheme:
            return None
        if category not in self.scheme[className]:
            return None

        output = 'Category: %s; scheme="%s"; class="%s"' % (
            category,
            self.scheme[className][category]["scheme"],
            className,
        )
        for attribute in ["location", "title"]:
            if attribute in self.scheme[className][category]:
                output += '; %s="%s"' % (attribute, self.scheme[className][category][attribute])

        return output + "\n"

    def createInstance(self, instanceID="", createPublicIP=True):
        """
        This creates a VM instance for the given boot image
        and creates a context script, taken the given parameters.
        Successful creation returns instance VM

        Boots a new node on the OpenStack server defined by self.endpointConfig. The
        'personality' of the node is done by self.imageConfig. Both variables are
        defined on initialization phase.

        The node name has the following format:
        <bootImageName><contextMethod><time>

        It boots the node. If IPpool is defined on the imageConfiguration, a floating
        IP is created and assigned to the node.

        :return: S_OK( ( nodeID, publicIP ) ) | S_ERROR
        """

        imageID = self.parameters.get("ImageID")
        flavor = self.parameters.get("FlavorName")
        self.parameters["VMUUID"] = instanceID
        self.parameters["VMType"] = self.parameters.get("CEType", "Occi")

        result = self._createUserDataScript()
        if not result["OK"]:
            return result
        userData = str(result["Value"])

        headers = {"Accept": "text/plain,text/occi", "Content-Type": "text/plain,text/occi", "Connection": "close"}

        data = self.__renderCategory("compute", "kind")
        for cat in ["user_data", imageID, flavor]:
            item = self.__renderCategory(cat, "mixin")
            if item is None:
                return S_ERROR("Category %s not defined in the scheme" % cat)
            data += item
        data += 'X-OCCI-Attribute: occi.core.id="%s"\n' % str(uuid.uuid4())
        data += 'X-OCCI-Attribute: occi.core.title="%s"\n' % instanceID
        data += 'X-OCCI-Attribute: occi.compute.hostname="%s"\n' % instanceID
        data += 'X-OCCI-Attribute: org.openstack.compute.user_data="%s"' % base64.b64encode(userData)
        # data += 'X-OCCI-Attribute: org.openstack.credentials.publickey.data="ssh-rsa ' + sshPublicKey + ' vmdirac"'

        del self.authArgs["data"]

        result = self.session.post(self.computeUrl, data=data, headers=headers, **self.authArgs)

        # print "AT >>> createInstance", result, result.headers
        # print "AT >>> result.text", result.text

        if result.status_code == 201:
            nodeID = result.text.split()[-1]
            return S_OK((nodeID, None))

        return S_ERROR("Failed VM creation: %s" % result.text)

    def getVMIDs(self):
        """Get all the VM IDs on the endpoint

        :return: list of VM ids
        """

        try:
            response = self.session.get(self.computeUrl)
        except Exception as e:
            return S_ERROR("Cannot connect to " + self.computeUrl + " (" + str(e) + ")")

        vmIDs = [id.split()[1] for id in response.text.split("\n") if id.startswith("X-OCCI-Location:")]
        return S_OK(vmIDs)

    def getVMStatus(self, nodeID):
        """
        Get the status for a given node ID. libcloud translates the status into a digit
        from 0 to 4 using a many-to-one relation ( ACTIVE and RUNNING -> 0 ), which
        means we cannot undo that translation. It uses an intermediate states mapping
        dictionary, SITEMAP, which we use here inverted to return the status as a
        meaningful string. The five possible states are ( ordered from 0 to 4 ):
        RUNNING, REBOOTING, TERMINATED, PENDING & UNKNOWN.

        :Parameters:
          **uniqueId** - `string`
            openstack node id ( not uuid ! )

        :return: S_OK( status ) | S_ERROR
        """
        url = "%s/%s" % (self.computeUrl, os.path.basename(nodeID))
        try:
            response = self.session.get(url)
        except Exception as e:
            return S_ERROR("Cannot get node details for %s (" % nodeID + str(e) + ")")

        status = "Unknown"
        for item in response.text.split("\n"):
            if "occi.compute.state" in item:
                status = item.split("=")[1].replace('"', "")
        return S_OK(status)

    def getVMNetworks(self):
        """Get a network object corresponding to the networkName

        :param str networkName: network name
        :return: S_OK|S_ERROR network object in case of S_OK
        """
        networkUrl = "%s/network/" % self.serviceUrl
        try:
            response = self.session.get(networkUrl)
        except Exception as e:
            return S_ERROR("Cannot get network details")

        if response.status_code != 200:
            return S_ERROR("Failed to get available networks")

        networks = []
        for line in response.text.split("\n"):
            networks.append(line.split()[1].split("/")[-1])

        return S_OK(networks)

    def getVMNetworkInterface(self, network):
        """Get a network object corresponding to the networkName

        :param str networkName: network name
        :return: S_OK|S_ERROR network object in case of S_OK
        """
        headers = {"Accept": "application/occi,application/json"}
        networkUrl = "%s/network/%s" % (self.serviceUrl, network)
        try:
            response = self.session.get(networkUrl)
        except Exception as e:
            return S_ERROR("Cannot get network details")

        if response.status_code != 200:
            return S_ERROR("Failed to get available networks")

        return S_OK(response.text)

    def stopVM(self, nodeID, publicIP=""):
        """
        Given the node ID it gets the node details, which are used to destroy the
        node making use of the libcloud.openstack driver. If three is any public IP
        ( floating IP ) assigned, frees it as well.

        :Parameters:
          **uniqueId** - `string`
            openstack node id ( not uuid ! )
          **public_ip** - `string`
            public IP assigned to the node if any

        :return: S_OK | S_ERROR
        """

        url = "%s/%s" % (self.computeUrl, os.path.basename(nodeID))
        try:
            response = self.session.delete(url)
        except Exception as e:
            return S_ERROR("Cannot delete node %s (" % nodeID + str(e) + ")")

        if response.status_code == 200:
            return S_OK(response.text)
        else:
            return S_ERROR(response.text)

    def assignFloatingIP(self, nodeID):
        """
        Given a node, assign a floating IP from the ipPool defined on the imageConfiguration
        on the CS.

        :Parameters:
          **node** - `libcloud.compute.base.Node`
            node object with the vm details

        :return: S_OK( public_ip ) | S_ERROR
        """

        result = self.getVMNetworks()
        if not result["OK"]:
            return result

        network = result["Value"][1]
        networkInterfaceID = str(uuid.uuid4())[:8]

        headers = {"Accept": "text/plain,text/occi", "Content-Type": "text/plain,text/occi", "Connection": "close"}

        nodeRef = "%s/%s" % (self.computeUrl, os.path.basename(nodeID))
        nodeRef = nodeRef.replace("//", "/")

        data = (
            'Category: networkinterface;scheme="http://schemas.ogf.org/occi/infrastructure#";'
            'class="kind";location="/link/networkinterface/";title="networkinterface link"\n'
        )
        data += 'X-OCCI-Attribute: occi.core.source="%s"\n' % nodeRef
        data += 'X-OCCI-Attribute: occi.core.target="%s/network/%s"\n' % (self.serviceUrl, network)
        data += 'X-OCCI-Attribute: occi.core.id="%s"' % networkInterfaceID

        headers["Content-Length"] = str(len(data))
        result = self.session.post("%s/link/networkinterface/" % self.serviceUrl, headers=headers, data=data)

        if result.status_code != 201:
            return S_ERROR(result.text)
        else:
            return S_OK(result.text.split()[1])

    def getVMFloatingIP(self, publicIP):

        return S_ERROR("Not implemented")

    def deleteFloatingIP(self, nodeID):
        """
        Deletes a floating IP <public_ip> from the server.

        :param str publicIP: public IP to be deleted
        :param object node: node to which IP is attached
        :return: S_OK | S_ERROR
        """

        headers = {"Accept": "text/plain,text/occi", "Content-Type": "text/occi", "Connection": "close"}

        nodeURL = "%s/link/networkinterface/%s" % (self.serviceUrl, os.path.basename(nodeID))
        nodeURL = nodeURL.replace("//", "/")

        result = self.session.delete(nodeURL, headers=headers)

        if result.status_code != 200:
            return S_ERROR(result.text)
        else:
            return S_OK(result.text)



"""
   Collection of DIRAC useful network related modules
   by default on Error they return None
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import socket
import six
from six.moves.urllib import parse as urlparse
import os
import struct
import array
import fcntl
import platform

from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR


def discoverInterfaces():
    max_possible = 128
    maxBytes = max_possible * 32
    mySocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    names = array.array("B", b"\0" * maxBytes)
    # 0x8912 SICGIFCONF
    fcntlOut = fcntl.ioctl(mySocket.fileno(), 0x8912, struct.pack("iL", maxBytes, names.buffer_info()[0]))
    outbytes = struct.unpack("iL", fcntlOut)[0]
    namestr = names.tobytes() if six.PY3 else names.tostring()  # pylint: disable=no-member

    if "32" in platform.architecture()[0]:
        step = 32
        offset = 32
    else:
        step = 40
        offset = 16

    ifaces = {}
    for i in range(0, outbytes, step):
        name = namestr[i : i + offset].split(b"\0", 1)[0].decode()
        ip = namestr[i + 20 : i + 24]
        ifaces[name] = {"ip": socket.inet_ntoa(ip), "mac": getMACFromInterface(name)}
    return ifaces


def getMACFromInterface(ifname):
    mySocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # bytearray is only needed here for Python 2
    info = bytearray(fcntl.ioctl(mySocket.fileno(), 0x8927, struct.pack("256s", ifname[:15].encode())))
    return "".join(["%02x:" % char for char in info[18:24]])[:-1]


def getFQDN():
    sFQDN = socket.getfqdn()
    if sFQDN.find("localhost") > -1:
        sFQDN = os.uname()[1]
        socket.getfqdn(sFQDN)
    return sFQDN


def splitURL(url):
    o = urlparse.urlparse(url)
    if o.scheme == "":
        return S_ERROR("'%s' URL is missing protocol" % url)
    path = o.path
    path = path.lstrip("/")
    return S_OK((o.scheme, o.hostname or "", o.port or 0, path))


def getIPsForHostName(hostName):
    try:
        ips = [t[4][0] for t in socket.getaddrinfo(hostName, 0)]
    except Exception as e:
        return S_ERROR("Can't get info for host %s: %s" % (hostName, str(e)))
    uniqueIPs = []
    for ip in ips:
        if ip not in uniqueIPs:
            uniqueIPs.append(ip)
    return S_OK(uniqueIPs)


def checkHostsMatch(host1, host2):
    ipLists = []
    for host in (host1, host2):
        result = getIPsForHostName(host)
        if not result["OK"]:
            return result
        ipLists.append(result["Value"])
    # Check
    for ip1 in ipLists[0]:
        if ip1 in ipLists[1]:
            return S_OK(True)
    return S_OK(False)



#!/usr/bin/env python
""" Virtual Machine Command Line Interface. """

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

# Define input on python2 in a forwards compatible way
if hasattr(__builtins__, "raw_input"):
    # pylint: disable=E0602,W0622
    input = raw_input

import pprint
import getpass

from DIRAC.Core.Base.CLI import CLI
from DIRAC.Core.Security.Locations import getProxyLocation
from DIRAC.Core.Utilities.PrettyPrint import printTable

from DIRAC.Resources.Cloud.EndpointFactory import EndpointFactory
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getPilotBootstrapParameters, getVMTypeConfig, getVMTypes
from DIRAC.Core.Utilities.File import makeGuid

__RCSID__ = "$Id$"


class VirtualMachineCLI(CLI):
    """Virtual Machine management console"""

    def __init__(self, vo=None):
        CLI.__init__(self)
        self.site = None
        self.endpoint = None
        self.project = None
        self.vmType = None
        self.vo = vo
        self.proxyLocation = None
        self.user = None
        self.password = None

    def do_connect(self, args):
        """Choose the specified cloud endpoint for connection

        usage:
          connect <site> [<endpoint> [project]]
        """

        self.site = None
        self.endpoint = None
        self.project = None

        argss = args.split()
        self.site = argss.pop(0)
        if argss:
            self.endpoint = argss.pop(0)
        if argss:
            self.project = argss.pop(0)

        result = getVMTypeConfig(self.site, self.endpoint)
        if not result["OK"]:
            print("ERROR: can not get the cloud endpoint configuration \n%s" % result["Message"])
            return
        ceDict = result["Value"]
        if not self.project:
            self.project = ceDict.get("Project")
        if not self.endpoint:
            self.endpoint = ceDict["CEName"]
        # Check for authentication details
        authType = ceDict.get("Auth")
        if authType and authType in ["x509", "voms"]:
            # We need proxy to proceed
            self.proxyLocation = None
            proxy = getProxyLocation()
            if not proxy:
                print("ERROR: Requested endpoint requires proxy but it is not found")
                return
            self.proxyLocation = proxy
        else:
            # We need user/login to proceed
            if not ceDict.get("User") or not ceDict.get("Password"):
                print("Endpoint requires user/password")
                self.user = input(["Login:"])
                self.password = getpass.getpass("Password:")

        print("Connection: site=%s, endpoint=%s, project=%s" % (self.site, self.endpoint, self.project))
        self.prompt = "%s/%s/%s> " % (self.site, self.endpoint, self.project)

    def __getCE(self):
        """Get cloud Endpoint object"""

        result = EndpointFactory().getCE(self.site, self.endpoint, self.vmType)
        if not result["OK"]:
            print(result["Message"])
            return
        ce = result["Value"]

        # Add extra parameters if any
        extraParams = {}
        if self.project:
            extraParams["Project"] = self.project
        if self.user:
            extraParams["User"] = self.user
        if self.password:
            extraParams["Password"] = self.password

        if extraParams:
            ce.setParameters(extraParams)
            ce.initialize()

        return ce

    def __checkConnection(self):
        """Check that connection details are provided"""
        if not self.site:
            return False
        return True

    def do_list(self, args):
        """Get IDs of VM instances"""

        if not self.__checkConnection():
            print("No connection defined")
            return

        ce = self.__getCE()

        result = ce.getVMIDs()
        if not result["OK"]:
            print("ERROR: %s" % result["Message"])
        else:
            print("\n".join(result["Value"]))

    def do_info(self, args):
        """Get VM status"""

        argss = args.split()
        if not argss:
            print(self.do_status.__doc__)
            return

        longOutput = False
        arg = argss.pop(0)
        if arg == "-l":
            longOutput = True
            vmID = argss.pop(0)
        else:
            vmID = arg

        ce = self.__getCE()
        result = ce.getVMInfo(vmID)

        if not result["OK"]:
            print("ERROR: %s" % result["Message"])
        else:
            pprint.pprint(result["Value"])

    def do_sites(self, args):
        """List available cloud sites"""

        result = getVMTypes()

        print(result)

        siteDict = result["Value"]
        records = []
        for site in siteDict:
            ceStart = True
            for ce in siteDict[site]:
                vmStart = True
                for vmType in siteDict[site][ce]["VMTypes"]:
                    flavor = siteDict[site][ce]["VMTypes"][vmType].get("FlavorName", "Unknown")
                    image = siteDict[site][ce]["VMTypes"][vmType].get(
                        "Image", siteDict[site][ce]["VMTypes"][vmType].get("ImageID", "Unknown")
                    )
                    if ceStart and vmStart:
                        records.append([site, ce, vmType, flavor, image])
                    elif ceStart:
                        records.append(["", "", vmType, flavor, image])
                    else:
                        records.append(["", ce, vmType, flavor, image])
                    vmStart = False
                ceStart = False

        fields = ["Site", "Endpoint", "VM Type", "Flavor", "Image"]
        printTable(fields, records)

    def do_status(self, args):
        """Get VM status"""

        argss = args.split()
        if not argss:
            print(self.do_status.__doc__)
            return
        vmID = argss[0]
        del argss[0]
        longOutput = False
        if argss and args[0] == "-l":
            longOutput = True

        ce = self.__getCE()
        result = ce.getVMStatus(vmID)

        if not result["OK"]:
            print("ERROR: %s" % result["Message"])
        else:
            print(result["Value"]["status"])

    def do_ip(self, args):
        """Assign IP"""

        argss = args.split()
        if not argss:
            return
        vmID = argss[0]

        ce = self.__getCE()
        result = ce.assignFloatingIP(vmID)

        if not result["OK"]:
            print("ERROR: %s" % result["Message"])
        else:
            print(result["Value"])

    def do_token(self, args):
        """Display the current Keystone token if any"""

        ce = self.__getCE()
        if getattr(ce, "token"):
            print(ce.token)
        else:
            print("No token available")

    def do_create(self, args):
        """Create VM at the connected site

        usage:
          create <VMType> [ExtraParam1=Value1 [ExtraParam2=Value2]]
        """

        argss = args.split()
        if not argss:
            print(self.do_create.__doc__)
            return
        self.vmType = argss.pop(0)
        extraParameters = {}
        while argss:
            key, value = argss.pop(0).split("=")
            extraParameters[key] = value

        result = getPilotBootstrapParameters(vo=self.vo)
        bootParameters = result["Value"]
        bootParameters.update(extraParameters)

        ce = self.__getCE()
        ce.setBootstrapParameters(bootParameters)

        diracVMID = makeGuid()[:8]
        result = ce.createInstance(diracVMID)

        if not result["OK"]:
            print("ERROR: %s" % result["Message"])
        else:
            print(result["Value"])

    def do_stop(self, args):
        """Stop VM"""

        argss = args.split()
        if not argss:
            print(self.do_stop.__doc__)
            return
        vmID = argss[0]

        ce = self.__getCE()
        result = ce.stopVM(vmID)

        if not result["OK"]:
            print("ERROR: %s" % result["Message"])
        else:
            print("VM stopped")



"""
It used used to create different plots.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from six import BytesIO
import errno
from DIRAC.Core.Utilities.Graphs import barGraph, lineGraph, pieGraph, qualityGraph, textGraph, histogram

from DIRAC import S_OK, S_ERROR


def checkMetadata(metadata):
    """
    :param dict metadata: it contains information which will used in the plot creation.
    """
    if "span" in metadata:
        granularity = metadata["span"]
        if "starttime" in metadata:
            metadata["starttime"] = metadata["starttime"] - metadata["starttime"] % granularity
        if "endtime" in metadata:
            metadata["endtime"] = metadata["endtime"] - metadata["endtime"] % granularity
    if "limit_labels" not in metadata:
        metadata["limit_labels"] = 9999999


def generateNoDataPlot(fileName, data, metadata):
    """
    Tis generate an image with a specific error message.

    :param str fileName: name of the file
    :param list data: data
    :param dict metadata: metadata information
    """
    try:
        with open(fileName, "wb") as fn:
            text = "No data for this selection for the plot: %s" % metadata["title"]
            textGraph(text, fn, metadata)
    except IOError as e:
        return S_ERROR(errno.EIO, e)
    return S_OK()


def generateErrorMessagePlot(msgText):
    """
    It creates a plot whith a specific error message

    :param str msgText: the text which will appear on the plot.
    :return: the plot.
    """
    fn = BytesIO()
    textGraph(msgText, fn, {})
    data = fn.getvalue()
    fn.close()
    return S_OK(data)


def generateTimedStackedBarPlot(fileName, data, metadata):
    """
    It is used to create a time based line plot.

    :param str fileName: the name of the file
    :param list data: the data which is used to create the plot
    :param dict metadata: extra information used to create the plot.
    """
    try:
        with open(fileName, "wb") as fn:
            checkMetadata(metadata)
            for key, value in (("sort_labels", "sum"), ("legend_unit", "%")):
                if key not in metadata:
                    metadata[key] = value
            barGraph(data, fn, **metadata)
    except IOError as e:
        return S_ERROR(errno.EIO, e)
    return S_OK()


def generateQualityPlot(fileName, data, metadata):
    """
    It is used to create 2D plots.

    :param str fileName: the name of the file
    :param list data: the data which is used to create the plot
    :param dict metadata: extra information used to create the plot.
    """
    try:
        with open(fileName, "wb") as fn:
            checkMetadata(metadata)
            metadata["legend"] = False
            # HACK: Pad a bit to the left until the proper padding is calculated
            maxKeyLength = max([len(key) for key in data])
            metadata["sort_labels"] = "alpha"
            metadata["plot_left_padding"] = int(maxKeyLength * 2.5)
            qualityGraph(data, fn, **metadata)
    except IOError as e:
        return S_ERROR(errno.EIO, e)
    return S_OK()


def generateCumulativePlot(fileName, data, metadata):
    """
    It is used to create cumulativ plots.

    :param str fileName: the name of the file
    :param list data: the data which is used to create the plot
    :param dict metadata: extra information used to create the plot.
    """
    try:
        with open(fileName, "wb") as fn:
            checkMetadata(metadata)
            if "sort_labels" not in metadata:
                metadata["sort_labels"] = "last_value"
            lineGraph(data, fn, **metadata)
    except IOError as e:
        return S_ERROR(errno.EIO, e)
    return S_OK()


def generateStackedLinePlot(fileName, data, metadata):
    """
    It is used to create stacked line plot.

    :param str fileName: the name of the file
    :param list data: the data which is used to create the plot
    :param dict metadata: extra information used to create the plot.
    """
    try:
        with open(fileName, "wb") as fn:
            checkMetadata(metadata)
            for key, value in (("sort_labels", "sum"), ("legend_unit", "%")):
                if key not in metadata:
                    metadata[key] = value
            lineGraph(data, fn, **metadata)
    except IOError as e:
        return S_ERROR(errno.EIO, e)
    return S_OK()


def generatePiePlot(fileName, data, metadata):
    """
    It is used to create pie charts.

    :param str fileName: the nanme of the file
    :param list data: the data which is used to create the plot
    :param dict metadata: extra information used to create the plot.
    """
    try:
        with open(fileName, "wb") as fn:
            checkMetadata(metadata)
            pieGraph(data, fn, **metadata)
    except IOError as e:
        return S_ERROR(errno.EIO, e)
    return S_OK()


def generateHistogram(filename, data, metadata):
    """
    It is used to create histograms.

    :param str fileName: the nanme of the file
    :param list data: the data which is used to create the plot
    :param dict metadata: extra information used to create the plot.
    """
    try:
        with open(filename, "wb") as fn:
            checkMetadata(metadata)
            histogram(data, fn, **metadata)
    except IOError as e:
        return S_ERROR(errno.EIO, e)
    return S_OK()



########################################################################
# File :   LocalComputingElement.py
# Author : Ricardo Graciani, A.T.
########################################################################

""" LocalComputingElement is a class to handle non-grid computing clusters

Allows direct submission to underlying Batch Systems.

**Configuration Parameters**

Configuration for the LocalComputingElement submission can be done via the configuration system.

BatchSystem:
   Underlying batch system that is going to be used to orchestrate executable files. The Batch System has to be
   accessible from the LocalCE. By default, the LocalComputingElement submits directly on the host via the Host class.

SharedArea:
   Area used to store executable/output/error files if they are not aready defined via BatchOutput, BatchError,
   InfoArea, ExecutableArea and/or WorkArea. The path should be absolute.

BatchOutput:
   Area where the job outputs are stored.
   If not defined: SharedArea + '/data' is used.
   If not absolute: SharedArea + path is used.

BatchError:
   Area where the job errors are stored.
   If not defined: SharedArea + '/data' is used.
   If not absolute: SharedArea + path is used.

ExecutableArea:
   Area where the executable files are stored if necessary.
   If not defined: SharedArea + '/data' is used.
   If not absolute: SharedArea + path is used.

**Code Documentation**

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import stat
import shutil
import tempfile
import getpass
import errno
from six.moves.urllib.parse import urlparse

from DIRAC import S_OK, S_ERROR
from DIRAC import gConfig

from DIRAC.Resources.Computing.ComputingElement import ComputingElement
from DIRAC.Resources.Computing.PilotBundle import bundleProxy, writeScript
from DIRAC.Core.Utilities.List import uniqueElements
from DIRAC.Core.Utilities.File import makeGuid
from DIRAC.Core.Utilities.Subprocess import systemCall


class LocalComputingElement(ComputingElement):
    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(LocalComputingElement, self).__init__(ceUniqueID)

        self.ceType = ""
        self.execution = "Local"
        self.submittedJobs = 0
        self.userName = getpass.getuser()

    def _reset(self):
        """Process CE parameters and make necessary adjustments"""
        batchSystemName = self.ceParameters.get("BatchSystem", "Host")
        result = self.loadBatchSystem(batchSystemName)
        if not result["OK"]:
            self.log.error("Failed to load the batch system plugin %s", batchSystemName)
            return result

        self.queue = self.ceParameters["Queue"]
        if "ExecQueue" not in self.ceParameters or not self.ceParameters["ExecQueue"]:
            self.ceParameters["ExecQueue"] = self.ceParameters.get("Queue", "")
        self.execQueue = self.ceParameters["ExecQueue"]
        self.log.info("Using queue: ", self.queue)

        self.sharedArea = self.ceParameters["SharedArea"]
        self.batchOutput = self.ceParameters["BatchOutput"]
        if not self.batchOutput.startswith("/"):
            self.batchOutput = os.path.join(self.sharedArea, self.batchOutput)
        self.batchError = self.ceParameters["BatchError"]
        if not self.batchError.startswith("/"):
            self.batchError = os.path.join(self.sharedArea, self.batchError)
        self.infoArea = self.ceParameters["InfoArea"]
        if not self.infoArea.startswith("/"):
            self.infoArea = os.path.join(self.sharedArea, self.infoArea)
        self.executableArea = self.ceParameters["ExecutableArea"]
        if not self.executableArea.startswith("/"):
            self.executableArea = os.path.join(self.sharedArea, self.executableArea)
        self.workArea = self.ceParameters["WorkArea"]
        if not self.workArea.startswith("/"):
            self.workArea = os.path.join(self.sharedArea, self.workArea)

        result = self._prepareHost()
        if not result["OK"]:
            self.log.error("Failed to initialize CE", self.ceName)
            return result

        self.removeOutput = True
        if "RemoveOutput" in self.ceParameters:
            if self.ceParameters["RemoveOutput"].lower() in ["no", "false", "0"]:
                self.removeOutput = False

        self.submitOptions = self.ceParameters.get("SubmitOptions", "")
        self.numberOfProcessors = self.ceParameters.get("NumberOfProcessors", 1)
        self.wholeNode = self.ceParameters.get("WholeNode", False)
        # numberOfNodes is treated as a string as it can contain values such as "2-4"
        # where 2 would represent the minimum number of nodes to allocate, and 4 the maximum
        self.numberOfNodes = self.ceParameters.get("NumberOfNodes", "1")
        self.numberOfGPUs = self.ceParameters.get("NumberOfGPUs")

        return S_OK()

    def _addCEConfigDefaults(self):
        """Method to make sure all necessary Configuration Parameters are defined"""
        # First assure that any global parameters are loaded
        ComputingElement._addCEConfigDefaults(self)
        # Now batch system specific ones
        if "ExecQueue" not in self.ceParameters:
            self.ceParameters["ExecQueue"] = self.ceParameters.get("Queue", "")

        if "SharedArea" not in self.ceParameters:
            defaultPath = os.environ.get("HOME", ".")
            self.ceParameters["SharedArea"] = gConfig.getValue("/LocalSite/InstancePath", defaultPath)

        if "BatchOutput" not in self.ceParameters:
            self.ceParameters["BatchOutput"] = "data"

        if "BatchError" not in self.ceParameters:
            self.ceParameters["BatchError"] = "data"

        if "ExecutableArea" not in self.ceParameters:
            self.ceParameters["ExecutableArea"] = "data"

        if "InfoArea" not in self.ceParameters:
            self.ceParameters["InfoArea"] = "info"

        if "WorkArea" not in self.ceParameters:
            self.ceParameters["WorkArea"] = "work"

    def _prepareHost(self):
        """Prepare directories and copy control script"""

        # Make remote directories
        dirTuple = uniqueElements(
            [self.sharedArea, self.executableArea, self.infoArea, self.batchOutput, self.batchError, self.workArea]
        )
        cmdTuple = ["mkdir", "-p"] + dirTuple
        self.log.verbose("Creating working directories")
        result = systemCall(30, cmdTuple)
        if not result["OK"]:
            self.log.error("Failed creating working directories", "(%s)" % result["Message"][1])
            return result
        status, output, error = result["Value"]
        if status != 0:
            self.log.error("Failed to create directories", "(%s)" % error)
            return S_ERROR(errno.EACCES, "Failed to create directories")

        return S_OK()

    def submitJob(self, executableFile, proxy=None, numberOfJobs=1):
        if not os.access(executableFile, 5):
            os.chmod(executableFile, stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)

        # if no proxy is supplied, the executable can be submitted directly
        # otherwise a wrapper script is needed to get the proxy to the execution node
        # The wrapper script makes debugging more complicated and thus it is
        # recommended to transfer a proxy inside the executable if possible.
        if self.proxy and not proxy:
            proxy = self.proxy
        if proxy:
            self.log.verbose("Setting up proxy for payload")
            wrapperContent = bundleProxy(executableFile, proxy)
            name = writeScript(wrapperContent, os.getcwd())
            submitFile = name
        else:  # no proxy
            submitFile = executableFile

        jobStamps = []
        for _i in range(numberOfJobs):
            jobStamps.append(makeGuid()[:8])

        batchDict = {
            "Executable": submitFile,
            "NJobs": numberOfJobs,
            "OutputDir": self.batchOutput,
            "ErrorDir": self.batchError,
            "SubmitOptions": self.submitOptions,
            "ExecutionContext": self.execution,
            "JobStamps": jobStamps,
            "Queue": self.queue,
            "WholeNode": self.wholeNode,
            "NumberOfProcessors": self.numberOfProcessors,
            "NumberOfNodes": self.numberOfNodes,
            "NumberOfGPUs": self.numberOfGPUs,
        }
        resultSubmit = self.batchSystem.submitJob(**batchDict)
        if proxy:
            os.remove(submitFile)

        if resultSubmit["Status"] == 0:
            self.submittedJobs += len(resultSubmit["Jobs"])
            # jobIDs = [ self.ceType.lower()+'://'+self.ceName+'/'+_id for _id in resultSubmit['Jobs'] ]
            # FIXME: It would be more proper to fix pilotCommands.__setFlavour where 'ssh' is hardcoded than
            # making this illogical fix, but there is no good way for pilotCommands to know its origin ceType.
            # So, the jobIDs here need to start with 'ssh', not ceType, to accomodate
            # them to those hardcoded in pilotCommands.__setFlavour
            batchSystemName = self.batchSystem.__class__.__name__.lower()
            jobIDs = ["ssh" + batchSystemName + "://" + self.ceName + "/" + _id for _id in resultSubmit["Jobs"]]
            result = S_OK(jobIDs)
            if "ExecutableToKeep" in resultSubmit:
                result["ExecutableToKeep"] = resultSubmit["ExecutableToKeep"]
        else:
            result = S_ERROR(resultSubmit["Message"])

        return result

    def killJob(self, jobIDList):
        """Kill a bunch of jobs"""

        batchDict = {"JobIDList": jobIDList, "Queue": self.queue}
        resultKill = self.batchSystem.killJob(**batchDict)
        if resultKill["Status"] == 0:
            return S_OK()
        return S_ERROR(resultKill["Message"])

    def getCEStatus(self):
        """Method to return information on running and pending jobs."""
        result = S_OK()
        result["SubmittedJobs"] = self.submittedJobs
        result["RunningJobs"] = 0
        result["WaitingJobs"] = 0

        batchDict = {"User": self.userName, "Queue": self.queue}
        resultGet = self.batchSystem.getCEStatus(**batchDict)
        if resultGet["Status"] == 0:
            result["RunningJobs"] = resultGet.get("Running", 0)
            result["WaitingJobs"] = resultGet.get("Waiting", 0)
        else:
            result = S_ERROR(resultGet["Message"])

        self.log.verbose("Waiting Jobs: ", result["WaitingJobs"])
        self.log.verbose("Running Jobs: ", result["RunningJobs"])

        return result

    def getJobStatus(self, jobIDList):
        """Get the status information for the given list of jobs"""
        resultDict = {}
        jobDict = {}

        # Extract the batch job ID from the full DIRAC job ID
        for job in jobIDList:
            stamp = os.path.basename(urlparse(job).path)
            jobDict[stamp] = job
        stampList = list(jobDict)

        # Get the status for a given batch job ID
        batchDict = {"JobIDList": stampList, "User": self.userName, "Queue": self.queue}
        resultGet = self.batchSystem.getJobStatus(**batchDict)

        if resultGet["Status"] != 0:
            return S_ERROR(resultGet["Message"])

        # Construct the dictionary to return: resultDict[dirac job ID] = status
        for stamp, status in resultGet["Jobs"].items():
            resultDict[jobDict[stamp]] = status

        return S_OK(resultDict)

    def getJobOutput(self, jobID, localDir=None):
        """Get the specified job standard output and error files. If the localDir is provided,
        the output is returned as file in this directory. Otherwise, the output is returned
        as strings.
        """
        self.log.verbose("Getting output for jobID", jobID)
        result = self._getJobOutputFiles(jobID)
        if not result["OK"]:
            return result

        jobStamp, _host, outputFile, errorFile = result["Value"]

        if not localDir:
            tempDir = tempfile.mkdtemp()
        else:
            tempDir = localDir

        try:
            localOut = os.path.join(tempDir, "%s.out" % jobStamp)
            localErr = os.path.join(tempDir, "%s.err" % jobStamp)
            if os.path.exists(outputFile):
                shutil.copy(outputFile, localOut)
            if os.path.exists(errorFile):
                shutil.copy(errorFile, localErr)
        except Exception as x:
            return S_ERROR("Failed to get output files: %s" % str(x))

        open(localOut, "a").close()
        open(localErr, "a").close()

        # The result is OK, we can remove the output
        if self.removeOutput and os.path.exists(outputFile):
            os.remove(outputFile)
        if self.removeOutput and os.path.exists(errorFile):
            os.remove(errorFile)

        if localDir:
            return S_OK((localOut, localErr))

        # Return the output as a string
        with open(localOut, "r") as outputFile:
            output = outputFile.read()
        with open(localErr, "r") as errorFile:
            error = errorFile.read()
        shutil.rmtree(tempDir)
        return S_OK((output, error))

    def _getJobOutputFiles(self, jobID):
        """Get output file names for the specific CE"""
        jobStamp = os.path.basename(urlparse(jobID).path)
        host = urlparse(jobID).hostname

        if hasattr(self.batchSystem, "getJobOutputFiles"):
            batchDict = {
                "JobIDList": [jobStamp],
                "OutputDir": self.batchOutput,
                "ErrorDir": self.batchError,
                "NumberOfNodes": self.numberOfNodes,
            }
            result = self.batchSystem.getJobOutputFiles(**batchDict)
            if result["Status"] != 0:
                return S_ERROR("Failed to get job output files: %s" % result["Message"])

            output = result["Jobs"][jobStamp]["Output"]
            error = result["Jobs"][jobStamp]["Error"]
        else:
            output = "%s/%s.out" % (self.batchOutput, jobStamp)
            error = "%s/%s.out" % (self.batchError, jobStamp)

        return S_OK((jobStamp, host, output, error))



""" JobPolicy encapsulates authorization rules for different groups
    with respect to job related operations
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Security import Properties
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import (
    getUsernameForDN,
    getGroupsForUser,
    getPropertiesForGroup,
    getUsersInGroup,
)

RIGHT_GET_JOB = "GetJob"
RIGHT_GET_INFO = "GetInfo"
RIGHT_GET_SANDBOX = "GetSandbox"
RIGHT_PUT_SANDBOX = "PutSandbox"
RIGHT_CHANGE_STATUS = "ChangeStatus"
RIGHT_DELETE = "Delete"
RIGHT_KILL = "Kill"
RIGHT_SUBMIT = "Submit"
RIGHT_RESCHEDULE = "Reschedule"
RIGHT_GET_STATS = "GetStats"
RIGHT_RESET = "Reset"

ALL_RIGHTS = [
    RIGHT_GET_JOB,
    RIGHT_GET_INFO,
    RIGHT_GET_SANDBOX,
    RIGHT_PUT_SANDBOX,
    RIGHT_CHANGE_STATUS,
    RIGHT_DELETE,
    RIGHT_KILL,
    RIGHT_SUBMIT,
    RIGHT_RESCHEDULE,
    RIGHT_GET_STATS,
    RIGHT_RESET,
]

ADMIN_RIGHTS = [
    RIGHT_GET_JOB,
    RIGHT_GET_INFO,
    RIGHT_GET_SANDBOX,
    RIGHT_PUT_SANDBOX,
    RIGHT_CHANGE_STATUS,
    RIGHT_DELETE,
    RIGHT_KILL,
    RIGHT_RESCHEDULE,
    RIGHT_GET_STATS,
    RIGHT_RESET,
]

OWNER_RIGHTS = [
    RIGHT_GET_INFO,
    RIGHT_GET_SANDBOX,
    RIGHT_PUT_SANDBOX,
    RIGHT_CHANGE_STATUS,
    RIGHT_DELETE,
    RIGHT_KILL,
    RIGHT_RESCHEDULE,
]

SHARED_GROUP_RIGHTS = OWNER_RIGHTS

# Rights with respect to non-owner's jobs or to newly created jobs
PROPERTY_RIGHTS = {}
PROPERTY_RIGHTS[Properties.JOB_ADMINISTRATOR] = ADMIN_RIGHTS
PROPERTY_RIGHTS[Properties.NORMAL_USER] = [RIGHT_SUBMIT]
PROPERTY_RIGHTS[Properties.GENERIC_PILOT] = [RIGHT_RESCHEDULE]
PROPERTY_RIGHTS[Properties.JOB_MONITOR] = [RIGHT_GET_INFO]


sLog = gLogger.getSubLogger(__name__)


class JobPolicy(object):
    def __init__(self, userDN, userGroup, allInfo=True):

        self.userDN = userDN
        self.userName = ""
        result = getUsernameForDN(userDN)
        if result["OK"]:
            self.userName = result["Value"]
        self.userGroup = userGroup
        self.userProperties = getPropertiesForGroup(userGroup, [])
        self.jobDB = None
        self.allInfo = allInfo
        self.__permissions = {}
        self.__getUserJobPolicy()

    def getUserRightsForJob(self, jobID, owner=None, group=None):
        """Get access rights to job with jobID for the user specified by
        userDN/userGroup
        """
        if owner is None or group is None:
            result = self.jobDB.getJobAttributes(jobID, ["Owner", "OwnerGroup"])
            if not result["OK"]:
                return result
            elif result["Value"]:
                owner = result["Value"]["OwnerDN"]
                group = result["Value"]["OwnerGroup"]
            else:
                return S_ERROR("Job not found")

        return self.getJobPolicy(owner, group)

    def __getUserJobPolicy(self):
        """Get the job rights for the primary user for which the JobPolicy object
        is created
        """
        # Can not do anything by default
        for right in ALL_RIGHTS:
            self.__permissions[right] = False

        # Anybody can get info about the jobs
        if self.allInfo:
            self.__permissions[RIGHT_GET_INFO] = True

        # Give JobAdmin permission if needed
        if Properties.JOB_ADMINISTRATOR in self.userProperties:
            for right in PROPERTY_RIGHTS[Properties.JOB_ADMINISTRATOR]:
                self.__permissions[right] = True

        # Give JobMonitor permission if needed
        if Properties.JOB_MONITOR in self.userProperties:
            for right in PROPERTY_RIGHTS[Properties.JOB_MONITOR]:
                self.__permissions[right] = True

        # Give normal user permission if needed
        if Properties.NORMAL_USER in self.userProperties:
            for right in PROPERTY_RIGHTS[Properties.NORMAL_USER]:
                self.__permissions[right] = True

        # Give permissions of the generic pilot
        if Properties.GENERIC_PILOT in self.userProperties:
            for right in PROPERTY_RIGHTS[Properties.GENERIC_PILOT]:
                self.__permissions[right] = True

    def getJobPolicy(self, jobOwner="", jobOwnerGroup=""):
        """Get the job operations rights for a job owned by jobOwnerDN/jobOwnerGroup
        for a user with userDN/userGroup.
        Returns a dictionary of various operations rights
        """
        permDict = dict(self.__permissions)
        # Job Owner can do everything with his jobs
        if jobOwner and self.userName and jobOwner == self.userName:
            for right in OWNER_RIGHTS:
                permDict[right] = True

        # Members of the same group sharing their jobs can do everything
        if jobOwnerGroup == self.userGroup:
            if Properties.JOB_SHARING in self.userProperties:
                for right in SHARED_GROUP_RIGHTS:
                    permDict[right] = True

        return S_OK(permDict)

    def evaluateJobRights(self, jobList, right):
        """Get access rights to jobID for the user ownerDN/ownerGroup"""
        validJobList = []
        invalidJobList = []
        nonauthJobList = []
        ownerJobList = []
        userRights = {}

        result = self.jobDB.getJobsAttributes(jobList, ["Owner", "OwnerGroup"])
        if not result["OK"]:
            sLog.error(
                "evaluateJobRights: failure while getJobsAttributes",
                "for %s : %s" % (",".join(str(j) for j in jobList), result["Message"]),
            )
            return validJobList, invalidJobList, nonauthJobList, ownerJobList
        jobDict = result["Value"]
        for jID in jobList:
            jobID = int(jID)
            if jobID not in jobDict:
                invalidJobList.append(jobID)
                continue
            owner = jobDict[jobID]["Owner"]
            group = jobDict[jobID]["OwnerGroup"]

            if (owner, group) in userRights:
                rightDict = userRights[(owner, group)]
            else:
                result = self.getUserRightsForJob(jobID, owner=owner, group=group)
                if result["OK"]:
                    rightDict = result["Value"]
                    userRights[(owner, group)] = rightDict
                else:
                    sLog.error(
                        "evaluateJobRights: failure while getUserRightsForJob",
                        "for %s : %s" % (str(jobID), result["Message"]),
                    )
                    invalidJobList.append(jobID)

            if rightDict[right]:
                validJobList.append(jobID)
            else:
                nonauthJobList.append(jobID)
            if owner == self.userName:
                ownerJobList.append(jobID)

        return validJobList, invalidJobList, nonauthJobList, ownerJobList

    def getControlledUsers(self, right):
        """Get users and groups which jobs are subject to the given access right"""

        userGroupList = "ALL"
        # If allInfo flag is defined we can see info for any job
        if right == RIGHT_GET_INFO and self.allInfo:
            return S_OK(userGroupList)

        # Administrators can do everything
        if Properties.JOB_ADMINISTRATOR in self.userProperties:
            return S_OK(userGroupList)

        # Inspectors can see info for all the jobs
        if Properties.JOB_MONITOR in self.userProperties and right == RIGHT_GET_INFO:
            return S_OK(userGroupList)

        userGroupList = []
        # User can do many things with his jobs
        if Properties.NORMAL_USER in self.userProperties and right in OWNER_RIGHTS:
            result = getGroupsForUser(self.userName)
            if not result["OK"]:
                return result
            groups = result["Value"]
            for group in groups:
                if "NormalUser" in getPropertiesForGroup(group, []):
                    userGroupList.append((self.userName, group))

        # User can do many things with the jobs in the shared group
        if Properties.JOB_SHARING in self.userProperties and right in SHARED_GROUP_RIGHTS:
            sharedUsers = getUsersInGroup(self.userGroup)
            for user in sharedUsers:
                userGroupList.append((user, self.userGroup))

        userGroupList = list(set(userGroupList))
        return S_OK(userGroupList)



"""
  DIRAC class to execute services

  In the most common case, DIRAC services are executed using the dirac-service command.
  dirac-service accepts a list positional arguments. These arguments have the form:
  [DIRAC System Name]/[DIRAC Service Name]
  dirac-service then:
  - produces a instance of ServiceReactor
  - loads the required modules using the ServiceReactor.loadAgentModules method
  - starts the execution loop using the ServiceReactor.serve() method

  Service modules must be placed under the Service directory of a DIRAC System.
  DIRAC Systems are called XXXSystem where XXX is the [DIRAC System Name], and
  must inherit from the base class RequestHandler

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import socket
import signal
import os
import sys
import multiprocessing

try:
    import selectors
except ImportError:
    import selectors2 as selectors

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.DISET.private.Service import Service
from DIRAC.Core.DISET.private.GatewayService import GatewayService
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Base.private.ModuleLoader import ModuleLoader
from DIRAC.Core.DISET.private.Protocols import gProtocolDict
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.ConfigurationSystem.Client import PathFinder


__RCSID__ = "$Id$"

#: Time during which the service does not accept new requests and handles those in the queue, if the backlog is too large
#: This sleep is repeated for as long as Service.wantsThrottle is truthy
THROTTLE_SERVICE_SLEEP_SECONDS = 0.25


class ServiceReactor(object):

    __transportExtraKeywords = {
        "SSLSessionTimeout": False,
        "IgnoreCRLs": False,
        "PacketTimeout": "timeout",
        "SocketBacklog": False,
    }

    def __init__(self):
        self.__services = {}
        self.__alive = True
        self.__loader = ModuleLoader("Service", PathFinder.getServiceSection, RequestHandler, moduleSuffix="Handler")
        self.__maxFD = 0
        self.__listeningConnections = {}
        self.__stats = ReactorStats()
        self.__processes = []

    def initialize(self, servicesList):
        try:
            servicesList.remove(GatewayService.GATEWAY_NAME)
            self.__services[GatewayService.GATEWAY_NAME] = GatewayService()
        except ValueError:
            # No GW in the service list
            pass

        result = self.__loader.loadModules(servicesList)
        if not result["OK"]:
            return result
        self.__serviceModules = self.__loader.getModules()

        for serviceName in self.__serviceModules:
            self.__services[serviceName] = Service(self.__serviceModules[serviceName])

        # Loop again to include the GW in case there is one (included in the __init__)
        for serviceName in self.__services:
            gLogger.info("Initializing %s" % serviceName)
            result = self.__services[serviceName].initialize()
            if not result["OK"]:
                return result
        return S_OK()

    def closeListeningConnections(self):
        gLogger.info("Closing listening connections...")
        for svcName in self.__listeningConnections:
            if "transport" in self.__listeningConnections[svcName]:
                try:
                    self.__listeningConnections[svcName]["transport"].close()
                except Exception:
                    pass
                del self.__listeningConnections[svcName]["transport"]
        gLogger.info("Connections closed")

    def __createListeners(self):
        for serviceName in self.__services:
            svcCfg = self.__services[serviceName].getConfig()
            protocol = svcCfg.getProtocol()
            port = svcCfg.getPort()
            if not port:
                return S_ERROR("No port defined for service %s" % serviceName)
            if protocol not in gProtocolDict:
                return S_ERROR("Protocol %s is not known for service %s" % (protocol, serviceName))
            self.__listeningConnections[serviceName] = {"port": port, "protocol": protocol}
            transportArgs = {}
            for kw in ServiceReactor.__transportExtraKeywords:
                value = svcCfg.getOption(kw)
                if value:
                    ikw = ServiceReactor.__transportExtraKeywords[kw]
                    if ikw:
                        kw = ikw
                    if kw == "timeout":
                        value = int(value)
                    transportArgs[kw] = value
            gLogger.verbose("Initializing %s transport" % protocol, svcCfg.getURL())
            transport = gProtocolDict[protocol]["transport"](("", port), bServerMode=True, **transportArgs)
            retVal = transport.initAsServer()
            if not retVal["OK"]:
                return S_ERROR(
                    "Cannot start listening connection for service %s: %s" % (serviceName, retVal["Message"])
                )
            self.__listeningConnections[serviceName]["transport"] = transport
            self.__listeningConnections[serviceName]["socket"] = transport.getSocket()
        return S_OK()

    def stopChildProcesses(self, _sig, frame):
        """
        It is used to properly stop the service when more than one process are used.
        In principle this is doing the job of runsv, becuase runsv only send a sigterm to the parent process...

        :param int _sig: the signal sent to the process
        :param object frame: execution frame which contains the child processes
        """

        handler = frame.f_locals.get("self")
        if handler and isinstance(handler, ServiceReactor):
            handler.stopAllProcess()

        for child in frame.f_locals.get("children", []):
            gLogger.info("Stopping child processes: %d" % child)
            os.kill(child, signal.SIGTERM)

        sys.exit(0)

    def serve(self):
        result = self.__createListeners()
        if not result["OK"]:
            self.__closeListeningConnections()
            return result
        for svcName in self.__listeningConnections:
            gLogger.always("Listening at %s" % self.__services[svcName].getConfig().getURL())

        isMultiProcessingAllowed = False
        for svcName in self.__listeningConnections:
            if self.__services[svcName].getConfig().getCloneProcesses() > 0:
                isMultiProcessingAllowed = True
                break
        if isMultiProcessingAllowed:
            signal.signal(signal.SIGTERM, self.stopChildProcesses)
            signal.signal(signal.SIGINT, self.stopChildProcesses)
            for svcName in self.__listeningConnections:
                clones = self.__services[svcName].getConfig().getCloneProcesses()
                for i in range(1, clones):
                    p = multiprocessing.Process(target=self.__startCloneProcess, args=(svcName, i))
                    self.__processes.append(p)
                    p.start()
                    gLogger.always("Started clone process %s for %s" % (i, svcName))

        while self.__alive:
            self.__acceptIncomingConnection()

    def stopAllProcess(self):
        """
        It stops all the running processes.
        """
        for process in self.__processes:
            gLogger.info("Stopping: PID=%d, name=%s, parentPid=%d" % (process.pid, process.name, process._parent_pid))
            if process.is_alive():
                process.terminate()
                self.__processes.remove(process)

    # This function runs in a different process
    def __startCloneProcess(self, svcName, i):
        self.__services[svcName].setCloneProcessId(i)
        self.__alive = i
        while self.__alive:
            self.__acceptIncomingConnection(svcName)

    def __getListeningSelector(self, svcName=False):
        sel = selectors.DefaultSelector()
        if svcName:
            sel.register(self.__listeningConnections[svcName]["socket"], selectors.EVENT_READ, svcName)
        else:
            for svcName, svcInfo in self.__listeningConnections.items():
                sel.register(svcInfo["socket"], selectors.EVENT_READ, svcName)
        return sel

    def __acceptIncomingConnection(self, svcName=False):
        """
        This method just gets the incoming connection, checks IP address
        and generates job. SSL/TLS handshake and execution of the remote call
        are made by Service._processInThread() (in another thread) so
        the service can accept other clients while another thread handling remote call

        :param str svcName=False: Name of a service if you use multiple
                                  services at the same time
        """
        sel = self.__getListeningSelector(svcName)
        while self.__alive:
            try:
                events = sel.select(timeout=10)
                if len(events) == 0:
                    return
                for key, event in events:
                    if event & selectors.EVENT_READ:
                        svcName = key.data
                        retVal = self.__listeningConnections[svcName]["transport"].acceptConnection()
                        if not retVal["OK"]:
                            gLogger.warn("Error while accepting a connection: ", retVal["Message"])
                            return
                        clientTransport = retVal["Value"]
            except socket.error:
                return
            self.__maxFD = max(self.__maxFD, clientTransport.oSocket.fileno())
            # Is it banned?
            clientIP = clientTransport.getRemoteAddress()[0]
            if clientIP in Registry.getBannedIPs():
                gLogger.warn("Client connected from banned ip %s" % clientIP)
                clientTransport.close()
                continue
            # Handle connection
            self.__stats.connectionStablished()
            self.__services[svcName].handleConnection(clientTransport)
            while self.__services[svcName].wantsThrottle:
                gLogger.warn("Sleeping as service requested throttling", svcName)
                time.sleep(THROTTLE_SERVICE_SLEEP_SECONDS)
            # Renew context?
            now = time.time()
            renewed = False
            for svcName in self.__listeningConnections:
                tr = self.__listeningConnections[svcName]["transport"]
                if now - tr.latestServerRenewTime() > self.__services[svcName].getConfig().getContextLifeTime():
                    result = tr.renewServerContext()
                    if result["OK"]:
                        renewed = True
            if renewed:
                sel = self.__getListeningSelector()

    def __closeListeningConnections(self):
        for svcName in self.__listeningConnections:
            lc = self.__listeningConnections[svcName]
            if "transport" in lc and lc["transport"]:
                lc["transport"].close()


class ReactorStats(object):
    def __init__(self):
        self.__connections = 0
        self.__established = 0
        self.__startTime = Time.dateTime()

    def connectionStablished(self):
        self.__connections += 1



#!/usr/bin/env python
"""
Script that facilitates the modification of a element through the command line.
However, the usage of this script will set the element token to the command
issuer with a duration of 1 day.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from datetime import datetime, timedelta

from DIRAC import gLogger, exit as DIRACExit, S_OK, version
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.ResourceStatusSystem.Client import ResourceStatusClient
from DIRAC.ResourceStatusSystem.PolicySystem import StateMachine
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations


subLogger = None


def registerSwitches():
    """
    Registers all switches that can be used while calling the script from the
    command line interface.
    """

    switches = (
        ("element=", "Element family to be Synchronized ( Site, Resource or Node )"),
        ("name=", "Name (or comma-separeted list of names) of the element where the change applies"),
        ("statusType=", "StatusType (or comma-separeted list of names), if none applies to all possible statusTypes"),
        ("status=", "Status to be changed"),
        ("reason=", "Reason to set the Status"),
        ("VO=", "VO to change a status for. When omitted, status will be changed for all VOs"),
    )

    for switch in switches:
        Script.registerSwitch("", switch[0], switch[1])


def registerUsageMessage():
    """
    Takes the script __doc__ and adds the DIRAC version to it
    """
    usageMessage = "  DIRAC %s\n" % version
    usageMessage += __doc__

    Script.setUsageMessage(usageMessage)


def parseSwitches():
    """
    Parses the arguments passed by the user
    """

    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs()
    if args:
        subLogger.error("Found the following positional args '%s', but we only accept switches" % args)
        subLogger.error("Please, check documentation below")
        Script.showHelp(exitCode=1)

    switches = dict(Script.getUnprocessedSwitches())
    switches.setdefault("statusType", None)
    switches.setdefault("VO", None)

    for key in ("element", "name", "status", "reason"):

        if key not in switches:
            subLogger.error("%s Switch missing" % key)
            subLogger.error("Please, check documentation below")
            Script.showHelp(exitCode=1)

    if not switches["element"] in ("Site", "Resource", "Node"):
        subLogger.error("Found %s as element switch" % switches["element"])
        subLogger.error("Please, check documentation below")
        Script.showHelp(exitCode=1)

    statuses = StateMachine.RSSMachine(None).getStates()

    if not switches["status"] in statuses:
        subLogger.error("Found %s as element switch" % switches["element"])
        subLogger.error("Please, check documentation below")
        Script.showHelp(exitCode=1)

    subLogger.debug("The switches used are:")
    map(subLogger.debug, switches.items())

    return switches


def checkStatusTypes(statusTypes):
    """
    To check if values for 'statusType' are valid
    """

    opsH = Operations().getValue("ResourceStatus/Config/StatusTypes/StorageElement")
    acceptableStatusTypes = opsH.replace(",", "").split()

    for statusType in statusTypes:
        if statusType not in acceptableStatusTypes and statusType != "all":
            acceptableStatusTypes.append("all")
            subLogger.error(
                "'%s' is a wrong value for switch 'statusType'.\n\tThe acceptable values are:\n\t%s"
                % (statusType, str(acceptableStatusTypes))
            )

    if "all" in statusType:
        return acceptableStatusTypes
    return statusTypes


def unpack(switchDict):
    """
    To split and process comma-separated list of values for 'name' and 'statusType'
    """

    switchDictSet = []
    names = []
    statusTypes = []

    if switchDict["name"] is not None:
        names = list(filter(None, switchDict["name"].split(",")))

    if switchDict["statusType"] is not None:
        statusTypes = list(filter(None, switchDict["statusType"].split(",")))
        statusTypes = checkStatusTypes(statusTypes)

    if len(names) > 0 and len(statusTypes) > 0:
        combinations = [(a, b) for a in names for b in statusTypes]
        for combination in combinations:
            n, s = combination
            switchDictClone = switchDict.copy()
            switchDictClone["name"] = n
            switchDictClone["statusType"] = s
            switchDictSet.append(switchDictClone)
    elif len(names) > 0 and len(statusTypes) == 0:
        for name in names:
            switchDictClone = switchDict.copy()
            switchDictClone["name"] = name
            switchDictSet.append(switchDictClone)
    elif len(names) == 0 and len(statusTypes) > 0:
        for statusType in statusTypes:
            switchDictClone = switchDict.copy()
            switchDictClone["statusType"] = statusType
            switchDictSet.append(switchDictClone)
    elif len(names) == 0 and len(statusTypes) == 0:
        switchDictClone = switchDict.copy()
        switchDictClone["name"] = None
        switchDictClone["statusType"] = None
        switchDictSet.append(switchDictClone)

    return switchDictSet


def getTokenOwner():
    """
    Function that gets the userName from the proxy
    """
    proxyInfo = getProxyInfo()
    if not proxyInfo["OK"]:
        return proxyInfo

    userName = proxyInfo["Value"]["username"]
    return S_OK(userName)


def setStatus(switchDict, tokenOwner):
    """
    Function that gets the user token, sets the validity for it. Gets the elements
    in the database for a given name and statusType(s). Then updates the status
    of all them adding a reason and the token.
    """

    rssClient = ResourceStatusClient.ResourceStatusClient()

    elements = rssClient.selectStatusElement(
        switchDict["element"],
        "Status",
        name=switchDict["name"],
        statusType=switchDict["statusType"],
        vO=switchDict["VO"],
        meta={"columns": ["Status", "StatusType"]},
    )

    if not elements["OK"]:
        return elements
    elements = elements["Value"]

    if not elements:
        subLogger.warn(
            "Nothing found for %s, %s, %s %s"
            % (switchDict["element"], switchDict["name"], switchDict["VO"], switchDict["statusType"])
        )
        return S_OK()

    tomorrow = datetime.utcnow().replace(microsecond=0) + timedelta(days=1)

    for status, statusType in elements:

        subLogger.debug("%s %s" % (status, statusType))

        if switchDict["status"] == status:
            subLogger.notice("Status for %s (%s) is already %s. Ignoring.." % (switchDict["name"], statusType, status))
            continue

        subLogger.debug(
            "About to set status %s -> %s for %s, statusType: %s, VO: %s, reason: %s"
            % (status, switchDict["status"], switchDict["name"], statusType, switchDict["VO"], switchDict["reason"])
        )
        result = rssClient.modifyStatusElement(
            switchDict["element"],
            "Status",
            name=switchDict["name"],
            statusType=statusType,
            status=switchDict["status"],
            reason=switchDict["reason"],
            vO=switchDict["VO"],
            tokenOwner=tokenOwner,
            tokenExpiration=tomorrow,
        )
        if not result["OK"]:
            return result

    return S_OK()


def run(switchDict):
    """
    Main function of the script
    """

    tokenOwner = getTokenOwner()
    if not tokenOwner["OK"]:
        subLogger.error(tokenOwner["Message"])
        DIRACExit(1)
    tokenOwner = tokenOwner["Value"]

    subLogger.notice("TokenOwner is %s" % tokenOwner)

    result = setStatus(switchDict, tokenOwner)
    if not result["OK"]:
        subLogger.error(result["Message"])
        DIRACExit(1)


@Script()
def main():
    global subLogger
    global registerUsageMessage

    subLogger = gLogger.getSubLogger(__file__)

    # Script initialization
    registerSwitches()
    registerUsageMessage()
    switchDict = parseSwitches()
    switchDictSets = unpack(switchDict)

    # Run script
    for switchDict in switchDictSets:
        run(switchDict)

    # Bye
    DIRACExit(0)


if __name__ == "__main__":
    main()
